#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

import json
from typing import Tuple, Any
from unittest.mock import AsyncMock, patch, Mock

import pytest
from pydantic import ValidationError

from neo4j_graphrag.exceptions import SchemaValidationError, SchemaExtractionError
from neo4j_graphrag.experimental.components.schema import (
    SchemaBuilder,
    NodeType,
    PropertyType,
    RelationshipType,
    ConstraintType,
    SchemaFromTextExtractor,
    GraphSchema,
    SchemaFromExistingGraphExtractor,
    Pattern,
)
import os
import tempfile
import yaml

from neo4j_graphrag.generation import PromptTemplate
from neo4j_graphrag.llm.types import LLMResponse
from neo4j_graphrag.utils.file_handler import FileFormat


def test_node_type_initialization_from_string() -> None:
    node_type = NodeType.model_validate("Label")
    assert isinstance(node_type, NodeType)
    assert node_type.label == "Label"
    assert len(node_type.properties) == 1
    assert node_type.properties[0].name == "name"
    assert node_type.properties[0].type == "STRING"


def test_node_type_additional_properties_default() -> None:
    # default behavior:
    node_type = NodeType.model_validate({"label": "Label"})
    assert node_type.additional_properties is True
    # Empty properties list now violates min_length=1 constraint
    with pytest.raises(ValidationError):
        NodeType.model_validate({"label": "Label", "properties": []})
    node_type = NodeType.model_validate(
        {"label": "Label", "properties": [{"name": "name", "type": "STRING"}]}
    )
    assert node_type.additional_properties is False

    # manually changing the default value
    # impossible cases: no properties and no additional
    with pytest.raises(ValidationError):
        NodeType.model_validate({"label": "Label", "additional_properties": False})
    with pytest.raises(ValidationError):
        NodeType.model_validate(
            {"label": "Label", "properties": [], "additional_properties": False}
        )

    # working case: properties and additional allowed
    node_type = NodeType.model_validate(
        {
            "label": "Label",
            "properties": [{"name": "name", "type": "STRING"}],
            "additional_properties": True,
        }
    )
    assert node_type.additional_properties is True


@pytest.mark.parametrize(
    "label",
    ["__Person", "Person__", "__Person__", "__KNOWS", "KNOWS__", "__KNOWS__"],
)
def test_node_type_rejects_dunder_label(label: str) -> None:
    with pytest.raises(ValidationError, match="reserved '__'"):
        NodeType.model_validate(
            {"label": label, "properties": [{"name": "name", "type": "STRING"}]}
        )


@pytest.mark.parametrize(
    "label",
    ["__KNOWS", "KNOWS__", "__KNOWS__"],
)
def test_relationship_type_rejects_dunder_label(label: str) -> None:
    with pytest.raises(ValidationError, match="reserved '__'"):
        RelationshipType.model_validate({"label": label})


def test_property_type_initalization() -> None:
    prop = PropertyType(name="email", type="STRING")
    assert prop.name == "email"
    assert prop.type == "STRING"
    assert prop.required is False


def test_property_type_with_required_true() -> None:
    prop = PropertyType(name="id", type="INTEGER", required=True)
    assert prop.required is True


def test_property_type_is_frozen() -> None:
    prop = PropertyType(name="email", type="STRING", required=False)

    with pytest.raises(ValidationError):
        prop.name = "other"

    with pytest.raises(ValidationError):
        prop.required = True


def test_relationship_type_initialization_from_string() -> None:
    relationship_type = RelationshipType.model_validate("REL")
    assert isinstance(relationship_type, RelationshipType)
    assert relationship_type.label == "REL"
    assert relationship_type.properties == []


def test_relationship_type_additional_properties_default() -> None:
    relationship_type = RelationshipType.model_validate({"label": "REL"})
    assert relationship_type.additional_properties is True
    relationship_type = RelationshipType.model_validate(
        {"label": "REL", "properties": []}
    )
    assert relationship_type.additional_properties is True
    relationship_type = RelationshipType.model_validate(
        {"label": "REL", "properties": [{"name": "name", "type": "STRING"}]}
    )
    assert relationship_type.additional_properties is False

    # manually changing the default value
    # auto-correction: no properties and additional_properties=False -> auto-corrected to True
    relationship_type = RelationshipType.model_validate(
        {"label": "REL", "additional_properties": False}
    )
    assert relationship_type.additional_properties is True  # auto-corrected
    relationship_type = RelationshipType.model_validate(
        {"label": "REL", "properties": [], "additional_properties": False}
    )
    assert relationship_type.additional_properties is True  # auto-corrected

    # working case: properties and additional allowed
    relationship_type = RelationshipType.model_validate(
        {
            "label": "REL",
            "properties": [{"name": "name", "type": "STRING"}],
            "additional_properties": True,
        }
    )
    assert relationship_type.additional_properties is True


def test_constraint_type_initialization() -> None:
    constraint = ConstraintType(
        type="UNIQUENESS", node_type="Person", property_name="name"
    )
    assert constraint.type == "UNIQUENESS"
    assert constraint.node_type == "Person"
    assert constraint.property_name == "name"


def test_constraint_type_is_frozen() -> None:
    constraint = ConstraintType(
        type="UNIQUENESS", node_type="Person", property_name="name"
    )

    with pytest.raises(ValidationError):
        constraint.type = "UNIQUENESS"

    with pytest.raises(ValidationError):
        constraint.node_type = "Organization"

    with pytest.raises(ValidationError):
        constraint.property_name = "id"


def test_schema_additional_node_types_default() -> None:
    schema_dict: dict[str, Any] = {
        "node_types": [],
    }
    schema = GraphSchema.model_validate(schema_dict)
    assert schema.additional_node_types is True

    schema_dict = {
        "node_types": ["Person"],
    }
    schema = GraphSchema.model_validate(schema_dict)
    assert schema.additional_node_types is False


def test_schema_additional_relationship_types_default() -> None:
    schema_dict: dict[str, Any] = {
        "node_types": [],
    }
    schema = GraphSchema.model_validate(schema_dict)
    assert schema.additional_relationship_types is True

    schema_dict = {
        "node_types": [],
        "relationship_types": ["REL"],
    }
    schema = GraphSchema.model_validate(schema_dict)
    assert schema.additional_relationship_types is False


def test_schema_additional_patterns_default() -> None:
    schema_dict: dict[str, Any] = {
        "node_types": [],
    }
    schema = GraphSchema.model_validate(schema_dict)
    assert schema.additional_patterns is True

    schema_dict = {
        "node_types": ["Person"],
        "relationship_types": ["REL"],
        "patterns": [("Person", "REL", "Person")],
    }
    schema = GraphSchema.model_validate(schema_dict)
    assert schema.additional_patterns is False


def test_schema_additional_parameter_validation() -> None:
    """Additional relationship types not allowed, but additional patterns allowed

    => raise Exception
    """
    schema_dict = {
        "node_types": [
            {
                "label": "Person",
                "properties": [
                    {
                        "name": "name",
                        "type": "STRING",
                    },
                    {"name": "height", "type": "INTEGER"},
                ],
            }
        ],
        "relationship_types": [
            {
                "label": "KNOWS",
            }
        ],
        "patterns": [
            ("Person", "KNOWS", "Person"),
        ],
        "additional_relationship_types": True,
        "additional_patterns": False,
    }
    with pytest.raises(
        ValidationError,
        match="`additional_relationship_types` must be set to False when using `additional_patterns=False`",
    ):
        GraphSchema.model_validate(schema_dict)


def test_schema_constraint_validation_property_not_in_node_type() -> None:
    schema_dict: dict[str, Any] = {
        "node_types": [
            {
                "label": "Person",
                "properties": [{"name": "name", "type": "STRING"}],
            }
        ],
        "constraints": [
            {"type": "UNIQUENESS", "node_type": "Person", "property_name": "email"}
        ],
    }

    with pytest.raises(SchemaValidationError) as exc_info:
        GraphSchema.model_validate(schema_dict)

    assert "Constraint references undefined property" in str(exc_info.value)
    assert "on node type 'Person'" in str(exc_info.value)


def test_schema_constraint_with_additional_properties_with_allows_unknown_property() -> (
    None
):
    # if additional_properties is True, we can define constraints that are not in the node_type
    schema_dict: dict[str, Any] = {
        "node_types": [
            {
                "label": "Person",
                "properties": [{"name": "name", "type": "STRING"}],
                "additional_properties": True,
            }
        ],
        "constraints": [
            {"type": "UNIQUENESS", "node_type": "Person", "property_name": "email"}
        ],
    }

    # Should raise - email is not allowed because the property is not defined in the node
    with pytest.raises(SchemaValidationError) as exc_info:
        GraphSchema.model_validate(schema_dict)

    assert "Constraint references undefined property 'email'" in str(exc_info.value)


def test_schema_with_valid_constraints() -> None:
    schema_dict: dict[str, Any] = {
        "node_types": [
            {"label": "Person", "properties": [{"name": "name", "type": "STRING"}]}
        ],
        "constraints": [
            {"type": "UNIQUENESS", "node_type": "Person", "property_name": "name"}
        ],
    }
    schema = GraphSchema.model_validate(schema_dict)

    assert len(schema.constraints) == 1
    assert schema.constraints[0].type == "UNIQUENESS"
    assert schema.constraints[0].node_type == "Person"
    assert schema.constraints[0].property_name == "name"


def test_schema_constraint_validation_invalid_node_type() -> None:
    schema_dict: dict[str, Any] = {
        "node_types": [
            {"label": "Person", "properties": [{"name": "name", "type": "STRING"}]}
        ],
        "constraints": [
            {
                "type": "UNIQUENESS",
                "node_type": "NonExistentNode",
                "property_name": "id",
            }
        ],
    }

    with pytest.raises(SchemaValidationError) as exc_info:
        GraphSchema.model_validate(schema_dict)

    assert "Constraint references undefined node type: NonExistentNode" in str(
        exc_info.value
    )


def test_schema_constraint_validation_missing_property_name() -> None:
    schema_dict: dict[str, Any] = {
        "node_types": [
            {"label": "Person", "properties": [{"name": "name", "type": "STRING"}]}
        ],
        "constraints": [
            {"type": "UNIQUENESS", "node_type": "Person", "property_name": ""}
        ],
    }

    with pytest.raises(SchemaValidationError) as exc_info:
        GraphSchema.model_validate(schema_dict)

    assert "Constraint has no property name" in str(exc_info.value)


@pytest.fixture
def valid_node_types() -> tuple[NodeType, ...]:
    return (
        NodeType(
            label="PERSON",
            description="An individual human being.",
            properties=[
                PropertyType(name="birth date", type="ZONED_DATETIME"),
                PropertyType(name="name", type="STRING", required=True),
            ],
            additional_properties=False,
        ),
        NodeType(
            label="ORGANIZATION",
            description="A structured group of people with a common purpose.",
            properties=[PropertyType(name="name", type="STRING")],
            additional_properties=True,
        ),
        NodeType(
            label="AGE",
            description="Age of a person in years.",
            properties=[PropertyType(name="value", type="INTEGER")],
            additional_properties=True,
        ),
    )


@pytest.fixture
def valid_relationship_types() -> tuple[RelationshipType, ...]:
    return (
        RelationshipType(
            label="EMPLOYED_BY",
            description="Indicates employment relationship.",
            properties=[
                PropertyType(name="start_time", type="LOCAL_DATETIME", required=True),
                PropertyType(name="end_time", type="LOCAL_DATETIME"),
            ],
            additional_properties=False,
        ),
        RelationshipType(
            label="ORGANIZED_BY",
            description="Indicates organization responsible for an event.",
        ),
        RelationshipType(
            label="ATTENDED_BY", description="Indicates attendance at an event."
        ),
    )


@pytest.fixture
def valid_patterns() -> tuple[Pattern, ...]:
    return (
        Pattern(source="PERSON", relationship="EMPLOYED_BY", target="ORGANIZATION"),
        Pattern(source="ORGANIZATION", relationship="ATTENDED_BY", target="PERSON"),
    )


@pytest.fixture
def patterns_with_invalid_entity() -> tuple[Pattern, ...]:
    return (
        Pattern(source="PERSON", relationship="EMPLOYED_BY", target="ORGANIZATION"),
        Pattern(
            source="NON_EXISTENT_ENTITY", relationship="ATTENDED_BY", target="PERSON"
        ),
    )


@pytest.fixture
def valid_constraints() -> tuple[ConstraintType, ...]:
    return (
        ConstraintType(type="UNIQUENESS", node_type="PERSON", property_name="name"),
    )


@pytest.fixture
def patterns_with_invalid_relation() -> tuple[Pattern, ...]:
    return (
        Pattern(
            source="PERSON", relationship="NON_EXISTENT_RELATION", target="ORGANIZATION"
        ),
    )


@pytest.fixture
def schema_builder() -> SchemaBuilder:
    return SchemaBuilder()


@pytest.fixture
def graph_schema(
    schema_builder: SchemaBuilder,
    valid_node_types: Tuple[NodeType, ...],
    valid_relationship_types: Tuple[RelationshipType, ...],
    valid_patterns: tuple[Pattern, ...],
) -> GraphSchema:
    return schema_builder.create_schema_model(
        list(valid_node_types), list(valid_relationship_types), list(valid_patterns)
    )


def test_create_schema_model_valid_data(
    schema_builder: SchemaBuilder,
    valid_node_types: Tuple[NodeType, ...],
    valid_relationship_types: Tuple[RelationshipType, ...],
    valid_patterns: tuple[Pattern, ...],
) -> None:
    schema = schema_builder.create_schema_model(
        list(valid_node_types), list(valid_relationship_types), list(valid_patterns)
    )

    assert schema.node_types == valid_node_types
    assert schema.relationship_types == valid_relationship_types
    assert schema.patterns == valid_patterns
    assert schema.additional_node_types is False
    assert schema.additional_relationship_types is False
    assert schema.additional_patterns is False


def test_create_schema_model_with_constraints(
    schema_builder: SchemaBuilder,
    valid_node_types: Tuple[NodeType, ...],
    valid_constraints: Tuple[ConstraintType, ...],
) -> None:
    schema = schema_builder.create_schema_model(
        list(valid_node_types),
        constraints=list(valid_constraints),
    )

    assert schema.node_types == valid_node_types
    assert schema.constraints == valid_constraints
    assert len(schema.constraints) == 1
    assert schema.constraints[0].type == "UNIQUENESS"
    assert schema.constraints[0].node_type == "PERSON"
    assert schema.constraints[0].property_name == "name"


@pytest.mark.asyncio
async def test_run_method(
    schema_builder: SchemaBuilder,
    valid_node_types: Tuple[NodeType, ...],
    valid_relationship_types: Tuple[RelationshipType, ...],
    valid_patterns: tuple[Pattern, ...],
) -> None:
    with patch.object(
        schema_builder,
        "create_schema_model",
        return_value=GraphSchema(
            node_types=valid_node_types,
            relationship_types=valid_relationship_types,
            patterns=valid_patterns,
        ),
    ):
        schema = await schema_builder.run(
            list(valid_node_types), list(valid_relationship_types), list(valid_patterns)
        )

    assert schema.node_types == valid_node_types
    assert schema.relationship_types == valid_relationship_types
    assert schema.patterns == valid_patterns
    assert schema.additional_node_types is False
    assert schema.additional_relationship_types is False
    assert schema.additional_patterns is False


@pytest.mark.asyncio
async def test_run_method_with_constraints(
    schema_builder: SchemaBuilder,
    valid_node_types: Tuple[NodeType, ...],
    valid_constraints: Tuple[ConstraintType, ...],
) -> None:
    schema = await schema_builder.run(
        list(valid_node_types),
        constraints=list(valid_constraints),
    )

    assert schema.node_types == valid_node_types
    assert schema.constraints == valid_constraints
    assert len(schema.constraints) == 1
    assert schema.constraints[0].type == "UNIQUENESS"
    assert schema.constraints[0].node_type == "PERSON"
    assert schema.constraints[0].property_name == "name"


def test_create_schema_model_invalid_entity(
    schema_builder: SchemaBuilder,
    valid_node_types: Tuple[NodeType, ...],
    valid_relationship_types: Tuple[RelationshipType, ...],
    patterns_with_invalid_entity: tuple[Pattern, ...],
) -> None:
    with pytest.raises(SchemaValidationError) as exc_info:
        schema_builder.create_schema_model(
            list(valid_node_types),
            list(valid_relationship_types),
            list(patterns_with_invalid_entity),
        )
    assert "Node type 'NON_EXISTENT_ENTITY' is not defined" in str(
        exc_info.value
    ), "Should fail due to non-existent entity"


def test_create_schema_model_invalid_relation(
    schema_builder: SchemaBuilder,
    valid_node_types: Tuple[NodeType, ...],
    valid_relationship_types: Tuple[RelationshipType, ...],
    patterns_with_invalid_relation: tuple[Pattern, ...],
) -> None:
    with pytest.raises(SchemaValidationError) as exc_info:
        schema_builder.create_schema_model(
            list(valid_node_types),
            list(valid_relationship_types),
            list(patterns_with_invalid_relation),
        )
    assert "Relationship type 'NON_EXISTENT_RELATION' is not defined" in str(
        exc_info.value
    ), "Should fail due to non-existent relation"


def test_create_schema_model_no_potential_schema(
    schema_builder: SchemaBuilder,
    valid_node_types: Tuple[NodeType, ...],
    valid_relationship_types: Tuple[RelationshipType, ...],
) -> None:
    schema_instance = schema_builder.create_schema_model(
        list(valid_node_types), list(valid_relationship_types)
    )
    assert schema_instance.node_types == valid_node_types
    assert schema_instance.relationship_types == valid_relationship_types
    assert schema_instance.patterns == ()


def test_create_schema_model_no_relations_or_potential_schema(
    schema_builder: SchemaBuilder,
    valid_node_types: Tuple[NodeType, ...],
) -> None:
    schema_instance = schema_builder.create_schema_model(list(valid_node_types))

    assert len(schema_instance.node_types) == 3
    person = schema_instance.node_type_from_label("PERSON")

    assert person is not None
    assert person.description == "An individual human being."
    assert len(person.properties) == 2
    assert person.additional_properties is False

    org = schema_instance.node_type_from_label("ORGANIZATION")
    assert org is not None
    assert org.description == "A structured group of people with a common purpose."
    assert org.additional_properties is True

    age = schema_instance.node_type_from_label("AGE")
    assert age is not None
    assert age.description == "Age of a person in years."
    assert age.additional_properties is True


def test_create_schema_model_missing_relations(
    schema_builder: SchemaBuilder,
    valid_node_types: Tuple[NodeType, ...],
    valid_patterns: tuple[Pattern, ...],
) -> None:
    with pytest.raises(SchemaValidationError) as exc_info:
        schema_builder.create_schema_model(
            node_types=valid_node_types, patterns=valid_patterns
        )
    assert "Relationship types must also be provided when using patterns." in str(
        exc_info.value
    ), "Should fail due to missing relations"


@pytest.fixture
def mock_llm() -> AsyncMock:
    mock = AsyncMock()
    mock.ainvoke = AsyncMock()
    return mock


@pytest.fixture
def valid_schema_json() -> str:
    return """
    {
        "node_types": [
            {
                "label": "Person",
                "properties": [
                    {"name": "name", "type": "STRING"}
                ]
            },
            {
                "label": "Organization",
                "properties": [
                    {"name": "name", "type": "STRING"}
                ]
            }
        ],
        "relationship_types": [
            {
                "label": "WORKS_FOR",
                "properties": [
                    {"name": "since", "type": "DATE"}
                ]
            }
        ],
        "patterns": [
            ["Person", "WORKS_FOR", "Organization"]
        ]
    }
    """


@pytest.fixture
def schema_json_with_valid_constraints() -> str:
    return """
    {
        "node_types": [
            {
                "label": "Person",
                "properties": [
                    {"name": "name", "type": "STRING"},
                    {"name": "email", "type": "STRING"}
                ]
            },
            {
                "label": "Organization",
                "properties": [
                    {"name": "name", "type": "STRING"}
                ]
            }
        ],
        "relationship_types": [
            {
                "label": "WORKS_FOR",
                "properties": [
                    {"name": "since", "type": "DATE"}
                ]
            }
        ],
        "patterns": [
            ["Person", "WORKS_FOR", "Organization"]
        ],
        "constraints": [
            {"type": "UNIQUENESS", "node_type": "Person", "property_name": "name"}
        ]
    }
    """


@pytest.fixture
def schema_json_with_invalid_constraints() -> str:
    return """
    {
        "node_types": [
            {
                "label": "Person",
                "properties": [
                    {"name": "name", "type": "STRING"}
                ]
            },
            {
                "label": "Organization",
                "properties": [
                    {"name": "name", "type": "STRING"}
                ]
            }
        ],
        "relationship_types": [
            {
                "label": "WORKS_FOR",
                "properties": [
                    {"name": "since", "type": "DATE"}
                ]
            }
        ],
        "patterns": [
            ["Person", "WORKS_FOR", "Organization"]
        ],
        "constraints": [
            {"type": "UNIQUENESS", "node_type": "Person", "property_name": "name"},
            {"type": "UNIQUENESS", "node_type": "Person", "property_name": "email"},
            {"type": "UNIQUENESS", "node_type": "NonExistentNode", "property_name": "id"},
            {"type": "UNIQUENESS", "node_type": "Person", "property_name": ""}
        ]
    }
    """


@pytest.fixture
def schema_json_with_null_constraints() -> str:
    return """
    {
        "node_types": [
            {
                "label": "Person",
                "properties": [
                    {"name": "name", "type": "STRING"}
                ]
            },
            {
                "label": "Organization",
                "properties": [
                    {"name": "name", "type": "STRING"}
                ]
            }
        ],
        "relationship_types": [
            {
                "label": "WORKS_FOR",
                "properties": [
                    {"name": "since", "type": "DATE"}
                ]
            }
        ],
        "patterns": [
            ["Person", "WORKS_FOR", "Organization"]
        ],
        "constraints": null
    }
    """


@pytest.fixture
def schema_json_with_required_properties() -> str:
    return """
    {
        "node_types": [
            {
                "label": "Person",
                "properties": [
                    {"name": "name", "type": "STRING", "required": true},
                    {"name": "email", "type": "STRING", "required": false},
                    {"name": "phone", "type": "STRING"}
                ]
            }
        ],
        "relationship_types": [
            {"label": "KNOWS"}
        ],
        "patterns": [
            ["Person", "KNOWS", "Person"]
        ]
    }
    """


@pytest.fixture
def schema_json_with_string_required_values() -> str:
    return """
    {
        "node_types": [
            {
                "label": "Person",
                "properties": [
                    {"name": "name", "type": "STRING", "required": "true"},
                    {"name": "email", "type": "STRING", "required": "yes"},
                    {"name": "phone", "type": "STRING", "required": "false"},
                    {"name": "address", "type": "STRING", "required": "no"}
                ]
            }
        ],
        "relationship_types": [
            {"label": "KNOWS"}
        ],
        "patterns": [
            ["Person", "KNOWS", "Person"]
        ]
    }
    """


@pytest.fixture
def invalid_schema_json() -> str:
    return """
    {
        "node_types": [
            {
                "label": "Person",
            },
        ],
        invalid json content
    }
    """


@pytest.fixture
def schema_from_text(mock_llm: AsyncMock) -> SchemaFromTextExtractor:
    return SchemaFromTextExtractor(llm=mock_llm)


@pytest.mark.asyncio
async def test_schema_from_text_run_valid_response(
    schema_from_text: SchemaFromTextExtractor,
    mock_llm: AsyncMock,
    valid_schema_json: str,
) -> None:
    # configure the mock LLM to return a valid schema JSON
    mock_llm.ainvoke.return_value = LLMResponse(content=valid_schema_json)

    # run the schema extraction
    schema_config = await schema_from_text.run(text="Sample text for extraction")

    # verify the LLM was called with a prompt
    mock_llm.ainvoke.assert_called_once()
    prompt_arg = mock_llm.ainvoke.call_args[0][0]
    assert isinstance(prompt_arg, str)
    assert "Sample text for extraction" in prompt_arg

    # verify the schema was correctly extracted
    assert len(schema_config.node_types) == 2
    assert schema_config.node_type_from_label("Person") is not None
    assert schema_config.node_type_from_label("Organization") is not None

    assert schema_config.relationship_types is not None
    assert schema_config.relationship_type_from_label("WORKS_FOR") is not None

    assert schema_config.patterns is not None
    assert len(schema_config.patterns) == 1
    assert schema_config.patterns[0] == ("Person", "WORKS_FOR", "Organization")


@pytest.mark.asyncio
async def test_schema_from_text_run_invalid_json(
    schema_from_text: SchemaFromTextExtractor,
    mock_llm: AsyncMock,
    invalid_schema_json: str,
) -> None:
    # configure the mock LLM to return invalid JSON
    mock_llm.ainvoke.return_value = LLMResponse(content=invalid_schema_json)

    # verify that running with invalid JSON raises a ValueError
    with pytest.raises(SchemaExtractionError) as exc_info:
        await schema_from_text.run(text="Sample text for extraction")

    assert "not valid JSON" in str(exc_info.value)


@pytest.mark.asyncio
async def test_schema_from_text_custom_template(
    mock_llm: AsyncMock, valid_schema_json: str
) -> None:
    # create a  custom template
    custom_prompt = "This is a custom prompt with text: {text}"
    custom_template = PromptTemplate(template=custom_prompt, expected_inputs=["text"])

    # create SchemaFromTextExtractor with the custom template
    schema_from_text = SchemaFromTextExtractor(
        llm=mock_llm, prompt_template=custom_template
    )

    # configure mock LLM to return valid JSON and capture the prompt that was sent to it
    mock_llm.ainvoke.return_value = LLMResponse(content=valid_schema_json)

    # run the schema extraction
    await schema_from_text.run(text="Sample text")

    # verify the custom prompt was passed to the LLM
    prompt_sent_to_llm = mock_llm.ainvoke.call_args[0][0]
    assert "This is a custom prompt with text" in prompt_sent_to_llm


@pytest.mark.asyncio
async def test_schema_from_text_llm_params(
    mock_llm: AsyncMock, valid_schema_json: str
) -> None:
    # configure custom LLM parameters
    llm_params = {"temperature": 0.1, "max_tokens": 500}

    # create SchemaFromTextExtractor with custom LLM parameters
    schema_from_text = SchemaFromTextExtractor(llm=mock_llm, llm_params=llm_params)

    # configure the mock LLM to return a valid schema JSON
    mock_llm.ainvoke.return_value = LLMResponse(content=valid_schema_json)

    # run the schema extraction
    await schema_from_text.run(text="Sample text")

    # verify the LLM was called with the custom parameters
    mock_llm.ainvoke.assert_called_once()
    call_kwargs = mock_llm.ainvoke.call_args[1]
    assert call_kwargs["temperature"] == 0.1
    assert call_kwargs["max_tokens"] == 500


@pytest.mark.asyncio
async def test_schema_config_save_json(graph_schema: GraphSchema) -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        # create file path
        json_path = os.path.join(temp_dir, "schema.json")

        # store the schema config
        graph_schema.save(json_path)

        # verify the file exists and has content
        assert os.path.exists(json_path)
        assert os.path.getsize(json_path) > 0

        # verify the content is valid JSON and contains expected data
        with open(json_path, "r") as f:
            data = json.load(f)
            assert "node_types" in data
            assert len(data["node_types"]) == 3


@pytest.mark.asyncio
async def test_schema_config_save_yaml(graph_schema: GraphSchema) -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        # Create file path
        yaml_path = os.path.join(temp_dir, "schema.yaml")

        # Store the schema config
        graph_schema.save(yaml_path)

        # Verify the file exists and has content
        assert os.path.exists(yaml_path)
        assert os.path.getsize(yaml_path) > 0

        # Verify the content is valid YAML and contains expected data
        with open(yaml_path, "r") as f:
            data = yaml.safe_load(f)
            assert "node_types" in data
            assert len(data["node_types"]) == 3


@pytest.mark.asyncio
async def test_schema_config_from_file(graph_schema: GraphSchema) -> None:
    with tempfile.TemporaryDirectory() as temp_dir:
        # create file paths with different extensions
        json_path = os.path.join(temp_dir, "schema.json")
        yaml_path = os.path.join(temp_dir, "schema.yaml")
        yml_path = os.path.join(temp_dir, "schema.yml")

        # store the schema config in the different formats
        graph_schema.save(json_path)
        graph_schema.save(yaml_path)
        graph_schema.save(yml_path)

        # load using from_file which should detect the format based on extension
        json_schema = GraphSchema.from_file(json_path)
        yaml_schema = GraphSchema.from_file(yaml_path)
        yml_schema = GraphSchema.from_file(yml_path)

        # simple verification that the objects were loaded correctly
        assert isinstance(json_schema, GraphSchema)
        assert isinstance(yaml_schema, GraphSchema)
        assert isinstance(yml_schema, GraphSchema)

        # verify basic structure is intact
        assert "node_types" in json_schema.model_dump()
        assert "node_types" in yaml_schema.model_dump()
        assert "node_types" in yml_schema.model_dump()

        # verify an unsupported extension raises the correct error
        txt_path = os.path.join(temp_dir, "schema.txt")
        graph_schema.save(
            txt_path, format=FileFormat.JSON
        )  # Store as JSON but with .txt extension

        with pytest.raises(ValueError, match="Unsupported file format: None"):
            GraphSchema.from_file(txt_path)


@pytest.fixture
def valid_schema_json_array() -> str:
    return """
    [
        {
            "node_types": [
                {
                    "label": "Person",
                    "properties": [
                        {"name": "name", "type": "STRING"}
                    ]
                },
                {
                    "label": "Organization",
                    "properties": [
                        {"name": "name", "type": "STRING"}
                    ]
                }
            ],
            "relationship_types": [
                {
                    "label": "WORKS_FOR",
                    "properties": [
                        {"name": "since", "type": "DATE"}
                    ]
                }
            ],
            "patterns": [
                ["Person", "WORKS_FOR", "Organization"]
            ]
        }
    ]
    """


@pytest.mark.asyncio
async def test_schema_from_text_run_valid_json_array(
    schema_from_text: SchemaFromTextExtractor,
    mock_llm: AsyncMock,
    valid_schema_json_array: str,
) -> None:
    # configure the mock LLM to return a valid JSON array
    mock_llm.ainvoke.return_value = LLMResponse(content=valid_schema_json_array)

    # run the schema extraction
    schema = await schema_from_text.run(text="Sample text for extraction")

    # verify the schema was correctly extracted from the array
    assert len(schema.node_types) == 2
    assert schema.node_type_from_label("Person") is not None
    assert schema.node_type_from_label("Organization") is not None

    assert schema.relationship_types is not None
    assert schema.relationship_type_from_label("WORKS_FOR") is not None

    assert schema.patterns is not None
    assert len(schema.patterns) == 1
    assert schema.patterns[0] == ("Person", "WORKS_FOR", "Organization")


@pytest.fixture
def schema_json_with_invalid_node_patterns() -> str:
    return """
    {
        "node_types": [
            {
                "label": "Person",
                "properties": [
                    {"name": "name", "type": "STRING"}
                ]
            },
            {
                "label": "Organization",
                "properties": [
                    {"name": "name", "type": "STRING"}
                ]
            }
        ],
        "relationship_types": [
            {
                "label": "WORKS_FOR",
                "properties": [
                    {"name": "since", "type": "DATE"}
                ]
            }
        ],
        "patterns": [
            ["Person", "WORKS_FOR", "Organization"],
            ["Person", "WORKS_FOR", "UndefinedNode"],
            ["UndefinedNode", "WORKS_FOR", "Organization"]
        ]
    }
    """


@pytest.fixture
def schema_json_with_invalid_relationship_patterns() -> str:
    return """
    {
        "node_types": [
            {
                "label": "Person",
                "properties": [
                    {"name": "name", "type": "STRING"}
                ]
            },
            {
                "label": "Organization",
                "properties": [
                    {"name": "name", "type": "STRING"}
                ]
            }
        ],
        "relationship_types": [
            {
                "label": "WORKS_FOR",
                "properties": [
                    {"name": "since", "type": "DATE"}
                ]
            }
        ],
        "patterns": [
            ["Person", "WORKS_FOR", "Organization"],
            ["Person", "UNDEFINED_RELATION", "Organization"],
            ["Organization", "ANOTHER_UNDEFINED_RELATION", "Person"]
        ]
    }
    """


@pytest.fixture
def schema_json_with_nodes_without_labels() -> str:
    return """
    {
        "node_types": [
            {
                "label": "Person",
                "properties": [
                    {"name": "name", "type": "STRING"}
                ]
            },
            {
                "properties": [
                    {"name": "name", "type": "STRING"}
                ]
            },
            {
                "label": "",
                "properties": [
                    {"name": "name", "type": "STRING"}
                ]
            },
            "Organization",
            "",
            "Company",
            "Invalid description with spaces",
            "{\\"invalid\\": \\"json object\\"}"
        ],
        "relationship_types": [
            {
                "label": "WORKS_FOR",
                "properties": [
                    {"name": "since", "type": "DATE"}
                ]
            }
        ],
        "patterns": [
            ["Person", "WORKS_FOR", "Organization"]
        ]
    }
    """


@pytest.fixture
def schema_json_with_relationships_without_labels() -> str:
    return """
    {
        "node_types": [
            {
                "label": "Person",
                "properties": [
                    {"name": "name", "type": "STRING"}
                ]
            },
            {
                "label": "Organization",
                "properties": [
                    {"name": "name", "type": "STRING"}
                ]
            }
        ],
        "relationship_types": [
            {
                "label": "WORKS_FOR",
                "properties": [
                    {"name": "since", "type": "DATE"}
                ]
            },
            {
                "properties": [
                    {"name": "since", "type": "DATE"}
                ]
            },
            {
                "label": "",
                "properties": [
                    {"name": "since", "type": "DATE"}
                ]
            },
            "MANAGES",
            "",
            "SUPERVISES",
            "invalid relationship description",
            "{\\"invalid\\": \\"json\\"}"
        ],
        "patterns": [
            ["Person", "WORKS_FOR", "Organization"],
            ["Person", "MANAGES", "Organization"]
        ]
    }
    """


@pytest.fixture
def schema_json_with_nonexistent_property_constraint() -> str:
    return """
    {
        "node_types": [
            {
                "label": "Person",
                "properties": [
                    {"name": "name", "type": "STRING"}
                ]
            }
        ],
        "relationship_types": [],
        "patterns": [],
        "constraints": [
            {"type": "UNIQUENESS", "node_type": "Person", "property_name": "name"},
            {"type": "UNIQUENESS", "node_type": "Person", "property_name": "nonexistent_property"}
        ]
    }
    """


@pytest.mark.asyncio
async def test_schema_from_text_filters_invalid_node_patterns(
    schema_from_text: SchemaFromTextExtractor,
    mock_llm: AsyncMock,
    schema_json_with_invalid_node_patterns: str,
) -> None:
    # configure the mock LLM to return schema with invalid node patterns
    mock_llm.ainvoke.return_value = LLMResponse(
        content=schema_json_with_invalid_node_patterns
    )

    # run the schema extraction
    schema = await schema_from_text.run(text="Sample text for extraction")

    # verify that invalid node patterns were filtered out (2 out of 3 patterns should be removed)
    assert schema.patterns is not None
    assert len(schema.patterns) == 1
    assert schema.patterns[0] == ("Person", "WORKS_FOR", "Organization")


@pytest.mark.asyncio
async def test_schema_from_text_filters_invalid_relationship_patterns(
    schema_from_text: SchemaFromTextExtractor,
    mock_llm: AsyncMock,
    schema_json_with_invalid_relationship_patterns: str,
) -> None:
    # configure the mock LLM to return schema with invalid relationship patterns
    mock_llm.ainvoke.return_value = LLMResponse(
        content=schema_json_with_invalid_relationship_patterns
    )

    # run the schema extraction
    schema = await schema_from_text.run(text="Sample text for extraction")

    # verify that invalid relationship patterns were filtered out (2 out of 3 patterns should be removed)
    assert schema.patterns is not None
    assert len(schema.patterns) == 1
    assert schema.patterns[0] == ("Person", "WORKS_FOR", "Organization")


@pytest.mark.asyncio
async def test_schema_from_text_filters_nodes_without_labels(
    schema_from_text: SchemaFromTextExtractor,
    mock_llm: AsyncMock,
    schema_json_with_nodes_without_labels: str,
) -> None:
    # configure the mock LLM to return schema with nodes without labels
    mock_llm.ainvoke.return_value = LLMResponse(
        content=schema_json_with_nodes_without_labels
    )

    # run the schema extraction
    schema = await schema_from_text.run(text="Sample text for extraction")

    # verify that nodes without labels were filtered out (5 out of 8 nodes should be removed)
    assert len(schema.node_types) == 3
    assert schema.node_type_from_label("Person") is not None
    assert schema.node_type_from_label("Organization") is not None
    assert schema.node_type_from_label("Company") is not None

    # verify that the pattern is still valid with the remaining nodes
    assert schema.patterns is not None
    assert len(schema.patterns) == 1
    assert schema.patterns[0] == ("Person", "WORKS_FOR", "Organization")


@pytest.mark.asyncio
async def test_schema_from_text_filters_relationships_without_labels(
    schema_from_text: SchemaFromTextExtractor,
    mock_llm: AsyncMock,
    schema_json_with_relationships_without_labels: str,
) -> None:
    # configure the mock LLM to return schema with relationships without labels
    mock_llm.ainvoke.return_value = LLMResponse(
        content=schema_json_with_relationships_without_labels
    )

    # run the schema extraction
    schema = await schema_from_text.run(text="Sample text for extraction")

    # verify that relationships without labels were filtered out (5 out of 8 relationships should be removed)
    assert schema.relationship_types is not None
    assert len(schema.relationship_types) == 3
    assert schema.relationship_type_from_label("WORKS_FOR") is not None
    assert schema.relationship_type_from_label("MANAGES") is not None
    assert schema.relationship_type_from_label("SUPERVISES") is not None

    # verify that the patterns are still valid with the remaining relationships
    assert schema.patterns is not None
    assert len(schema.patterns) == 2
    assert ("Person", "WORKS_FOR", "Organization") in schema.patterns
    assert ("Person", "MANAGES", "Organization") in schema.patterns


@pytest.mark.asyncio
async def test_schema_from_text_with_valid_constraints(
    schema_from_text: SchemaFromTextExtractor,
    mock_llm: AsyncMock,
    schema_json_with_valid_constraints: str,
) -> None:
    # configure the mock LLM to return schema with valid constraints
    mock_llm.ainvoke.return_value = LLMResponse(
        content=schema_json_with_valid_constraints
    )

    # run the schema extraction
    schema = await schema_from_text.run(text="Sample text for extraction")

    assert len(schema.constraints) == 1
    assert schema.constraints[0].type == "UNIQUENESS"
    assert schema.constraints[0].node_type == "Person"
    assert schema.constraints[0].property_name == "name"


@pytest.mark.asyncio
async def test_schema_from_text_filters_invalid_constraints(
    schema_from_text: SchemaFromTextExtractor,
    mock_llm: AsyncMock,
    schema_json_with_invalid_constraints: str,
) -> None:
    # configure the mock LLM to return schema with invalid constraints
    mock_llm.ainvoke.return_value = LLMResponse(
        content=schema_json_with_invalid_constraints
    )

    # run the schema extraction
    schema = await schema_from_text.run(text="Sample text for extraction")

    # verify that invalid constraints were filtered out:
    # constraints with NonExistentNode should be removed
    # constraint with empty property_name should be removed
    # only the valid constraint should remain
    assert len(schema.constraints) == 1
    assert schema.constraints[0].node_type == "Person"
    assert schema.constraints[0].property_name == "name"


@pytest.mark.asyncio
async def test_schema_from_text_filters_constraint_with_nonexistent_property(
    schema_from_text: SchemaFromTextExtractor,
    mock_llm: AsyncMock,
    schema_json_with_nonexistent_property_constraint: str,
) -> None:
    # configure the mock LLM to return schema with constraint on nonexistent property
    mock_llm.ainvoke.return_value = LLMResponse(
        content=schema_json_with_nonexistent_property_constraint
    )

    # run the schema extraction
    schema = await schema_from_text.run(text="Sample text for extraction")

    # verify that only the valid constraint (with "name" property) remains
    # the constraint with "nonexistent_property" should be filtered out
    assert len(schema.constraints) == 1
    assert schema.constraints[0].property_name == "name"


@pytest.mark.asyncio
async def test_schema_from_text_handles_null_constraints(
    schema_from_text: SchemaFromTextExtractor,
    mock_llm: AsyncMock,
    schema_json_with_null_constraints: str,
) -> None:
    # configure the mock LLM to return schema with null constraints
    mock_llm.ainvoke.return_value = LLMResponse(
        content=schema_json_with_null_constraints
    )

    # run the schema extraction - should not crash
    schema = await schema_from_text.run(text="Sample text for extraction")

    # verify schema was created with empty constraints
    assert len(schema.constraints) == 0


@pytest.mark.asyncio
async def test_schema_from_text_handles_missing_constraints(
    schema_from_text: SchemaFromTextExtractor,
    mock_llm: AsyncMock,
    valid_schema_json: str,
) -> None:
    # configure the mock LLM to return schema without constraints field
    mock_llm.ainvoke.return_value = LLMResponse(content=valid_schema_json)

    #  run the schema extraction - should not crash
    schema = await schema_from_text.run(text="Sample text for extraction")

    # verify schema was created with empty constraints
    assert len(schema.constraints) == 0


def test_clean_json_content_markdown_with_json_language(
    schema_from_text: SchemaFromTextExtractor,
) -> None:
    content = """```json
{"node_types": [{"label": "Person"}]}
```"""

    cleaned = schema_from_text._clean_json_content(content)
    assert cleaned == '{"node_types": [{"label": "Person"}]}'


def test_clean_json_content_markdown_without_language(
    schema_from_text: SchemaFromTextExtractor,
) -> None:
    content = """```
{"node_types": [{"label": "Person"}]}
```"""

    cleaned = schema_from_text._clean_json_content(content)
    assert cleaned == '{"node_types": [{"label": "Person"}]}'


def test_clean_json_content_plain_json(
    schema_from_text: SchemaFromTextExtractor,
) -> None:
    content = '{"node_types": [{"label": "Person"}]}'

    cleaned = schema_from_text._clean_json_content(content)
    assert cleaned == '{"node_types": [{"label": "Person"}]}'


def test_filter_properties_required_field_valid_true(
    schema_from_text: SchemaFromTextExtractor,
) -> None:
    node_types = [
        {
            "label": "Person",
            "properties": [{"name": "name", "type": "STRING", "required": True}],
        }
    ]
    result = schema_from_text._filter_properties_required_field(node_types)
    assert result[0]["properties"][0]["required"] is True


def test_filter_properties_required_field_valid_false(
    schema_from_text: SchemaFromTextExtractor,
) -> None:
    node_types = [
        {
            "label": "Person",
            "properties": [{"name": "name", "type": "STRING", "required": False}],
        }
    ]
    result = schema_from_text._filter_properties_required_field(node_types)
    assert result[0]["properties"][0]["required"] is False


def test_filter_properties_required_field_string(
    schema_from_text: SchemaFromTextExtractor,
) -> None:
    node_types = [
        {
            "label": "Person",
            "properties": [
                {"name": "prop1", "type": "STRING", "required": "true"},
                {"name": "prop2", "type": "STRING", "required": "yes"},
                {"name": "prop3", "type": "STRING", "required": "1"},
                {"name": "prop4", "type": "STRING", "required": "TRUE"},
            ],
        }
    ]
    result = schema_from_text._filter_properties_required_field(node_types)
    for prop in result[0]["properties"]:
        assert prop["required"] is True
    node_types = [
        {
            "label": "Person",
            "properties": [
                {"name": "prop1", "type": "STRING", "required": "false"},
                {"name": "prop2", "type": "STRING", "required": "no"},
                {"name": "prop3", "type": "STRING", "required": "0"},
                {"name": "prop4", "type": "STRING", "required": "FALSE"},
            ],
        }
    ]
    result = schema_from_text._filter_properties_required_field(node_types)
    for prop in result[0]["properties"]:
        assert prop["required"] is False


def test_filter_properties_required_field_invalid_string(
    schema_from_text: SchemaFromTextExtractor,
) -> None:
    node_types = [
        {
            "label": "Person",
            "properties": [
                {"name": "name", "type": "STRING", "required": "mandatory"},
                {"name": "email", "type": "STRING", "required": "always"},
            ],
        }
    ]
    result = schema_from_text._filter_properties_required_field(node_types)

    assert "required" not in result[0]["properties"][0]
    assert "required" not in result[0]["properties"][1]


def test_filter_properties_required_field_int_values(
    schema_from_text: SchemaFromTextExtractor,
) -> None:
    """Test that int values like 1 and 0 are converted to True/False."""
    node_types = [
        {
            "label": "Person",
            "properties": [
                {"name": "prop1", "type": "STRING", "required": 1},
                {"name": "prop2", "type": "STRING", "required": 0},
            ],
        }
    ]
    result = schema_from_text._filter_properties_required_field(node_types)
    assert result[0]["properties"][0]["required"] is True
    assert result[0]["properties"][1]["required"] is False


def test_filter_properties_required_field_invalid_type(
    schema_from_text: SchemaFromTextExtractor,
) -> None:
    """Test that unrecognized types like list and dict are removed."""
    node_types = [
        {
            "label": "Person",
            "properties": [
                {"name": "prop1", "type": "STRING", "required": []},
                {"name": "prop2", "type": "STRING", "required": {"value": True}},
            ],
        }
    ]
    result = schema_from_text._filter_properties_required_field(node_types)
    for prop in result[0]["properties"]:
        assert "required" not in prop


def test_filter_properties_required_field_missing(
    schema_from_text: SchemaFromTextExtractor,
) -> None:
    node_types = [
        {
            "label": "Person",
            "properties": [{"name": "name", "type": "STRING"}],
        }
    ]
    result = schema_from_text._filter_properties_required_field(node_types)
    assert "required" not in result[0]["properties"][0]


def test_enforce_required_for_constraint_properties_sets_required_true(
    schema_from_text: SchemaFromTextExtractor,
) -> None:
    node_types: list[dict[str, Any]] = [
        {
            "label": "Person",
            "properties": [
                {"name": "name", "type": "STRING", "required": False},
                {"name": "email", "type": "STRING", "required": False},
            ],
        }
    ]
    constraints = [
        {"type": "UNIQUENESS", "node_type": "Person", "property_name": "name"}
    ]

    schema_from_text._enforce_required_for_constraint_properties(
        node_types, constraints
    )

    # name should now be required=true
    assert node_types[0]["properties"][0]["required"] is True
    # email should remain required=false
    assert node_types[0]["properties"][1]["required"] is False


def test_enforce_required_for_constraint_properties_already_true(
    schema_from_text: SchemaFromTextExtractor,
) -> None:
    node_types: list[dict[str, Any]] = [
        {
            "label": "Person",
            "properties": [
                {"name": "name", "type": "STRING", "required": True},
            ],
        }
    ]
    constraints = [
        {"type": "UNIQUENESS", "node_type": "Person", "property_name": "name"}
    ]

    schema_from_text._enforce_required_for_constraint_properties(
        node_types, constraints
    )

    assert node_types[0]["properties"][0]["required"] is True


def test_enforce_required_for_constraint_properties_missing_required_field(
    schema_from_text: SchemaFromTextExtractor,
) -> None:
    node_types: list[dict[str, Any]] = [
        {
            "label": "Person",
            "properties": [
                {"name": "name", "type": "STRING"},  # No required field
            ],
        }
    ]
    constraints = [
        {"type": "UNIQUENESS", "node_type": "Person", "property_name": "name"}
    ]

    schema_from_text._enforce_required_for_constraint_properties(
        node_types, constraints
    )

    assert node_types[0]["properties"][0]["required"] is True


def test_enforce_required_for_constraint_properties_no_constraints(
    schema_from_text: SchemaFromTextExtractor,
) -> None:
    node_types: list[dict[str, Any]] = [
        {
            "label": "Person",
            "properties": [
                {"name": "name", "type": "STRING", "required": False},
            ],
        }
    ]
    constraints: list[dict[str, Any]] = []

    schema_from_text._enforce_required_for_constraint_properties(
        node_types, constraints
    )

    assert node_types[0]["properties"][0]["required"] is False


def test_enforce_required_for_constraint_properties_skips_unconstrained_nodes(
    schema_from_text: SchemaFromTextExtractor,
) -> None:
    node_types: list[dict[str, Any]] = [
        {
            "label": "Person",
            "properties": [
                {"name": "name", "type": "STRING", "required": False},
            ],
        },
        {
            "label": "Company",
            "properties": [
                {"name": "name", "type": "STRING", "required": False},
            ],
        },
    ]
    constraints = [
        {"type": "UNIQUENESS", "node_type": "Person", "property_name": "name"}
    ]

    schema_from_text._enforce_required_for_constraint_properties(
        node_types, constraints
    )

    # Person.name should be required=true
    assert node_types[0]["properties"][0]["required"] is True
    # Company.name should remain required=false (no constraint on Company)
    assert node_types[1]["properties"][0]["required"] is False


@pytest.mark.asyncio
async def test_schema_from_text_with_required_properties(
    schema_from_text: SchemaFromTextExtractor,
    mock_llm: AsyncMock,
    schema_json_with_required_properties: str,
) -> None:
    mock_llm.ainvoke.return_value = LLMResponse(
        content=schema_json_with_required_properties
    )

    schema = await schema_from_text.run(text="Sample text for test")

    person = schema.node_type_from_label("Person")
    assert person is not None

    # Check required properties
    name_prop = next((p for p in person.properties if p.name == "name"), None)
    email_prop = next((p for p in person.properties if p.name == "email"), None)
    phone_prop = next((p for p in person.properties if p.name == "phone"), None)

    assert name_prop is not None and name_prop.required is True
    assert email_prop is not None and email_prop.required is False
    assert phone_prop is not None and phone_prop.required is False


@pytest.mark.asyncio
async def test_schema_from_text_sanitizes_string_required_values(
    schema_from_text: SchemaFromTextExtractor,
    mock_llm: AsyncMock,
    schema_json_with_string_required_values: str,
) -> None:
    mock_llm.ainvoke.return_value = LLMResponse(
        content=schema_json_with_string_required_values
    )

    schema = await schema_from_text.run(text="Sample text for test")

    person = schema.node_type_from_label("Person")
    assert person is not None

    # true and yes should become True
    name_prop = next((p for p in person.properties if p.name == "name"), None)
    email_prop = next((p for p in person.properties if p.name == "email"), None)
    assert name_prop is not None and name_prop.required is True
    assert email_prop is not None and email_prop.required is True

    # false and no should become False
    phone_prop = next((p for p in person.properties if p.name == "phone"), None)
    address_prop = next((p for p in person.properties if p.name == "address"), None)
    assert phone_prop is not None and phone_prop.required is False
    assert address_prop is not None and address_prop.required is False


@pytest.mark.asyncio
async def test_schema_from_text_handles_missing_required_field(
    schema_from_text: SchemaFromTextExtractor,
    mock_llm: AsyncMock,
    valid_schema_json: str,
) -> None:
    mock_llm.ainvoke.return_value = LLMResponse(content=valid_schema_json)

    schema = await schema_from_text.run(text="Sample text")

    person = schema.node_type_from_label("Person")
    assert person is not None

    # All properties should have required=False (default)
    for prop in person.properties:
        assert prop.required is False


@pytest.mark.asyncio
async def test_schema_from_text_enforces_required_for_constrained_properties(
    schema_from_text: SchemaFromTextExtractor,
    mock_llm: AsyncMock,
) -> None:
    schema_json = """
    {
        "node_types": [
            {
                "label": "Person",
                "properties": [
                    {"name": "name", "type": "STRING", "required": false},
                    {"name": "email", "type": "STRING", "required": false}
                ]
            }
        ],
        "relationship_types": [],
        "patterns": [],
        "constraints": [
            {"type": "UNIQUENESS", "node_type": "Person", "property_name": "name"}
        ]
    }
    """
    mock_llm.ainvoke.return_value = LLMResponse(content=schema_json)

    schema = await schema_from_text.run(text="Sample text")

    person = schema.node_type_from_label("Person")
    assert person is not None

    name_prop = next((p for p in person.properties if p.name == "name"), None)
    email_prop = next((p for p in person.properties if p.name == "email"), None)

    # name should be auto-fixed to required=true
    assert name_prop is not None and name_prop.required is True
    # email should remain required=false
    assert email_prop is not None and email_prop.required is False


@pytest.mark.asyncio
@patch("neo4j_graphrag.experimental.components.schema.get_structured_schema")
async def test_schema_from_existing_graph(mock_get_structured_schema: Mock) -> None:
    mock_get_structured_schema.return_value = {
        "node_props": {
            "Person": [
                {"property": "id", "type": "INTEGER"},
                {"property": "name", "type": "STRING"},
            ]
        },
        "rel_props": {"KNOWS": [{"property": "fromDate", "type": "DATE"}]},
        "relationships": [
            {"start": "Person", "type": "KNOWS", "end": "Person"},
            {"start": "Person", "type": "LIVES_IN", "end": "City"},
        ],
        "metadata": {
            "constraint": [
                {
                    "id": 7,
                    "name": "person_id",
                    "type": "NODE_PROPERTY_EXISTENCE",
                    "entityType": "NODE",
                    "labelsOrTypes": ["Person"],
                    "properties": ["id"],
                    "ownedIndex": "person_id",
                    "propertyType": None,
                },
            ],
            "index": [
                {
                    "label": "Person",
                    "properties": ["name"],
                    "size": 2,
                    "type": "RANGE",
                    "valuesSelectivity": 1.0,
                    "distinctValues": 2.0,
                },
            ],
        },
    }
    driver = Mock()
    schema_builder = SchemaFromExistingGraphExtractor(
        driver=driver,
    )
    schema = await schema_builder.run()
    assert isinstance(schema, GraphSchema)
    assert len(schema.node_types) == 2
    person_node_type = schema.node_type_from_label("Person")
    assert person_node_type is not None
    id_person_property = [p for p in person_node_type.properties if p.name == "id"][0]
    assert id_person_property.required is True
    assert person_node_type.additional_properties is False
    city_node_type = schema.node_type_from_label("City")
    assert city_node_type is not None
    assert city_node_type.additional_properties is True

    assert len(schema.relationship_types) == 2
    knows_rel = schema.relationship_type_from_label("KNOWS")
    assert knows_rel is not None
    assert knows_rel.additional_properties is False
    lives_rel = schema.relationship_type_from_label("LIVES_IN")
    assert lives_rel is not None
    assert lives_rel.additional_properties is True

    assert schema.additional_node_types is False
    assert schema.additional_relationship_types is False
    assert schema.additional_patterns is False

    assert schema.patterns == (
        Pattern(source="Person", relationship="KNOWS", target="Person"),
        Pattern(source="Person", relationship="LIVES_IN", target="City"),
    )


@pytest.mark.asyncio
@patch("neo4j_graphrag.experimental.components.schema.get_structured_schema")
async def test_schema_from_existing_graph_additional_params(
    mock_get_structured_schema: Mock,
) -> None:
    mock_get_structured_schema.return_value = {
        "node_props": {
            "Person": [
                {"property": "id", "type": "INTEGER"},
                {"property": "name", "type": "STRING"},
            ]
        },
        "rel_props": {"KNOWS": [{"property": "fromDate", "type": "DATE"}]},
        "relationships": [
            {"start": "Person", "type": "KNOWS", "end": "Person"},
            {"start": "Person", "type": "LIVES_IN", "end": "City"},
        ],
    }
    driver = Mock()
    schema_builder = SchemaFromExistingGraphExtractor(
        driver=driver,
        additional_node_types=True,
        additional_relationship_types=True,
        additional_patterns=True,
        additional_properties=True,
    )
    schema = await schema_builder.run()
    person_node_type = schema.node_type_from_label("Person")
    assert person_node_type is not None
    assert person_node_type.additional_properties is True
    city_node_type = schema.node_type_from_label("City")
    assert city_node_type is not None
    assert city_node_type.additional_properties is True

    knows_rel = schema.relationship_type_from_label("KNOWS")
    assert knows_rel is not None
    assert knows_rel.additional_properties is True
    lives_rel = schema.relationship_type_from_label("LIVES_IN")
    assert lives_rel is not None
    assert lives_rel.additional_properties is True

    assert schema.additional_node_types is True
    assert schema.additional_relationship_types is True
    assert schema.additional_patterns is True



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Any, Optional
from unittest.mock import AsyncMock, MagicMock, Mock, patch
from typing import List

import httpx
import pytest
from neo4j_graphrag.exceptions import LLMGenerationError
from neo4j_graphrag.llm import LLMResponse, MistralAILLM
from neo4j_graphrag.types import LLMMessage
from neo4j_graphrag.utils.rate_limit import NoOpRateLimitHandler
from pydantic import BaseModel, ConfigDict


# Mock SDKError for testing
class MockSDKError(Exception):
    """Mock SDKError for testing purposes."""

    def __init__(
        self, message: str, raw_response: Optional[httpx.Response] = None
    ) -> None:
        super().__init__(message)
        self.raw_response = raw_response


@patch("neo4j_graphrag.llm.mistralai_llm.Mistral", None)
def test_mistralai_llm_missing_dependency() -> None:
    with pytest.raises(ImportError):
        MistralAILLM(model_name="mistral-model")


@patch("neo4j_graphrag.llm.mistralai_llm.Mistral")
def test_mistralai_llm_invoke(mock_mistral: Mock) -> None:
    mock_mistral_instance = mock_mistral.return_value

    chat_response_mock = MagicMock()
    chat_response_mock.choices = [
        MagicMock(message=MagicMock(content="mistral response"))
    ]

    mock_mistral_instance.chat.complete.return_value = chat_response_mock

    llm = MistralAILLM(model_name="mistral-model")

    res = llm.invoke("some input")

    assert isinstance(res, LLMResponse)
    assert res.content == "mistral response"


@patch("neo4j_graphrag.llm.mistralai_llm.Mistral")
def test_mistralai_llm_invoke_with_message_history(mock_mistral: Mock) -> None:
    mock_mistral_instance = mock_mistral.return_value
    chat_response_mock = MagicMock()
    chat_response_mock.choices = [
        MagicMock(message=MagicMock(content="mistral response"))
    ]
    mock_mistral_instance.chat.complete.return_value = chat_response_mock
    model = "mistral-model"
    system_instruction = "You are a helpful assistant."

    llm = MistralAILLM(model_name=model)

    message_history: List[LLMMessage] = [
        {"role": "user", "content": "When does the sun come up in the summer?"},
        {"role": "assistant", "content": "Usually around 6am."},
    ]
    question = "What about next season?"
    res = llm.invoke(question, message_history, system_instruction=system_instruction)

    assert isinstance(res, LLMResponse)
    assert res.content == "mistral response"
    messages: List[LLMMessage] = [{"role": "system", "content": system_instruction}]
    messages.extend(message_history)
    messages.append({"role": "user", "content": question})
    llm.client.chat.complete.assert_called_once_with(  # type: ignore[attr-defined]
        messages=messages,
        model=model,
    )


@patch("neo4j_graphrag.llm.mistralai_llm.Mistral")
def test_mistralai_llm_invoke_with_message_history_and_system_instruction(
    mock_mistral: Mock,
) -> None:
    mock_mistral_instance = mock_mistral.return_value
    chat_response_mock = MagicMock()
    chat_response_mock.choices = [
        MagicMock(message=MagicMock(content="mistral response"))
    ]
    mock_mistral_instance.chat.complete.return_value = chat_response_mock
    model = "mistral-model"
    system_instruction = "You are a helpful assistant."
    llm = MistralAILLM(model_name=model)
    message_history: List[LLMMessage] = [
        {"role": "user", "content": "When does the sun come up in the summer?"},
        {"role": "assistant", "content": "Usually around 6am."},
    ]
    question = "What about next season?"

    # first invocation - initial instructions
    res = llm.invoke(question, message_history, system_instruction=system_instruction)
    assert isinstance(res, LLMResponse)
    assert res.content == "mistral response"
    messages: List[LLMMessage] = [{"role": "system", "content": system_instruction}]
    messages.extend(message_history)
    messages.append({"role": "user", "content": question})
    llm.client.chat.complete.assert_called_once_with(  # type: ignore[attr-defined]
        messages=messages,
        model=model,
    )

    assert llm.client.chat.complete.call_count == 1  # type: ignore


@patch("neo4j_graphrag.llm.mistralai_llm.Mistral")
def test_mistralai_llm_invoke_with_message_history_validation_error(
    mock_mistral: Mock,
) -> None:
    mock_mistral_instance = mock_mistral.return_value
    chat_response_mock = MagicMock()
    chat_response_mock.choices = [
        MagicMock(message=MagicMock(content="mistral response"))
    ]
    mock_mistral_instance.chat.complete.return_value = chat_response_mock
    model = "mistral-model"
    system_instruction = "You are a helpful assistant."

    llm = MistralAILLM(model_name=model, system_instruction=system_instruction)

    message_history = [
        {"role": "user", "content": "When does the sun come up in the summer?"},
        {"role": "monkey", "content": "Usually around 6am."},
    ]
    question = "What about next season?"

    with pytest.raises(LLMGenerationError) as exc_info:
        llm.invoke(question, message_history)  # type: ignore
    assert "Input should be 'user', 'assistant' or 'system" in str(exc_info.value)


@pytest.mark.asyncio
@patch("neo4j_graphrag.llm.mistralai_llm.Mistral")
async def test_mistralai_llm_ainvoke(mock_mistral: Mock) -> None:
    mock_mistral_instance = mock_mistral.return_value

    async def mock_complete_async(*_args: Any, **_kwargs: Any) -> MagicMock:
        chat_response_mock = MagicMock()
        chat_response_mock.choices = [
            MagicMock(message=MagicMock(content="async mistral response"))
        ]
        return chat_response_mock

    mock_mistral_instance.chat.complete_async = mock_complete_async

    llm = MistralAILLM(model_name="mistral-model")

    res = await llm.ainvoke("some input")

    assert isinstance(res, LLMResponse)
    assert res.content == "async mistral response"


@patch("neo4j_graphrag.llm.mistralai_llm.SDKError", MockSDKError)
@patch("neo4j_graphrag.llm.mistralai_llm.Mistral")
def test_mistralai_llm_invoke_sdkerror(mock_mistral: Mock) -> None:
    mock_mistral_instance = mock_mistral.return_value
    raw_response = httpx.Response(status_code=500)
    mock_mistral_instance.chat.complete.side_effect = MockSDKError(
        "Some error", raw_response=raw_response
    )

    llm = MistralAILLM(model_name="mistral-model")

    with pytest.raises(LLMGenerationError):
        llm.invoke("some input")


@pytest.mark.asyncio
@patch("neo4j_graphrag.llm.mistralai_llm.SDKError", MockSDKError)
@patch("neo4j_graphrag.llm.mistralai_llm.Mistral")
async def test_mistralai_llm_ainvoke_sdkerror(mock_mistral: Mock) -> None:
    mock_mistral_instance = mock_mistral.return_value

    async def mock_complete_async(*args: Any, **kwargs: Any) -> None:
        raw_response = httpx.Response(status_code=500)
        raise MockSDKError("Some async error", raw_response=raw_response)

    mock_mistral_instance.chat.complete_async = mock_complete_async

    llm = MistralAILLM(model_name="mistral-model")

    with pytest.raises(LLMGenerationError):
        await llm.ainvoke("some input")


# V2 Interface Tests (List[LLMMessage] input)


@patch("neo4j_graphrag.llm.mistralai_llm.Mistral")
def test_mistralai_llm_invoke_v2_happy_path(mock_mistral: Mock) -> None:
    """Test V2 interface invoke method with List[LLMMessage] input."""
    mock_mistral_instance = mock_mistral.return_value
    chat_response_mock = MagicMock()
    chat_response_mock.choices = [
        MagicMock(message=MagicMock(content="mistral v2 response"))
    ]
    mock_mistral_instance.chat.complete.return_value = chat_response_mock

    messages: List[LLMMessage] = [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": "What is machine learning?"},
    ]

    llm = MistralAILLM(model_name="mistral-model")
    response = llm.invoke(messages)

    assert isinstance(response, LLMResponse)
    assert response.content == "mistral v2 response"

    # Verify the correct method was called
    llm.client.chat.complete.assert_called_once()  # type: ignore[attr-defined]
    call_args = llm.client.chat.complete.call_args[1]  # type: ignore[attr-defined]
    assert call_args["model"] == "mistral-model"
    assert len(call_args["messages"]) == 2


@patch("neo4j_graphrag.llm.mistralai_llm.Mistral")
def test_mistralai_llm_invoke_v2_with_conversation_history(mock_mistral: Mock) -> None:
    """Test V2 interface invoke method with complex conversation history."""
    mock_mistral_instance = mock_mistral.return_value
    chat_response_mock = MagicMock()
    chat_response_mock.choices = [
        MagicMock(message=MagicMock(content="mistral conversation response"))
    ]
    mock_mistral_instance.chat.complete.return_value = chat_response_mock

    messages: List[LLMMessage] = [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": "Tell me about Python."},
        {"role": "assistant", "content": "Python is a programming language."},
        {"role": "user", "content": "What about its history?"},
    ]

    llm = MistralAILLM(model_name="mistral-model")
    response = llm.invoke(messages)

    assert isinstance(response, LLMResponse)
    assert response.content == "mistral conversation response"

    # Verify the correct number of messages were passed
    llm.client.chat.complete.assert_called_once()  # type: ignore[attr-defined]
    call_args = llm.client.chat.complete.call_args[1]  # type: ignore[attr-defined]
    assert len(call_args["messages"]) == 4


@patch("neo4j_graphrag.llm.mistralai_llm.Mistral")
def test_mistralai_llm_invoke_v2_no_system_message(mock_mistral: Mock) -> None:
    """Test V2 interface invoke method without system message."""
    mock_mistral_instance = mock_mistral.return_value
    chat_response_mock = MagicMock()
    chat_response_mock.choices = [
        MagicMock(message=MagicMock(content="mistral no system response"))
    ]
    mock_mistral_instance.chat.complete.return_value = chat_response_mock

    messages: List[LLMMessage] = [
        {"role": "user", "content": "What is the capital of France?"},
    ]

    llm = MistralAILLM(model_name="mistral-model")
    response = llm.invoke(messages)

    assert isinstance(response, LLMResponse)
    assert response.content == "mistral no system response"

    # Verify only user message was passed
    llm.client.chat.complete.assert_called_once()  # type: ignore[attr-defined]
    call_args = llm.client.chat.complete.call_args[1]  # type: ignore[attr-defined]
    assert len(call_args["messages"]) == 1


@pytest.mark.asyncio
@patch("neo4j_graphrag.llm.mistralai_llm.Mistral")
async def test_mistralai_llm_ainvoke_v2_happy_path(mock_mistral: Mock) -> None:
    """Test V2 interface async invoke method with List[LLMMessage] input."""
    mock_mistral_instance = mock_mistral.return_value

    async def mock_complete_async(*_args: Any, **_kwargs: Any) -> MagicMock:
        chat_response_mock = MagicMock()
        chat_response_mock.choices = [
            MagicMock(message=MagicMock(content="async mistral v2 response"))
        ]
        return chat_response_mock

    mock_mistral_instance.chat.complete_async = mock_complete_async

    messages: List[LLMMessage] = [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": "What is async programming?"},
    ]

    llm = MistralAILLM(model_name="mistral-model")
    response = await llm.ainvoke(messages)

    assert isinstance(response, LLMResponse)
    assert response.content == "async mistral v2 response"


@pytest.mark.asyncio
@patch("neo4j_graphrag.llm.mistralai_llm.SDKError", MockSDKError)
@patch("neo4j_graphrag.llm.mistralai_llm.Mistral")
async def test_mistralai_llm_ainvoke_v2_error_handling(mock_mistral: Mock) -> None:
    """Test V2 interface async invoke method error handling."""
    mock_mistral_instance = mock_mistral.return_value

    async def mock_complete_async(*args: Any, **kwargs: Any) -> None:
        raise MockSDKError("V2 async error")

    mock_mistral_instance.chat.complete_async = mock_complete_async

    messages: List[LLMMessage] = [
        {"role": "user", "content": "This should fail"},
    ]

    llm = MistralAILLM(model_name="mistral-model")

    with pytest.raises(LLMGenerationError):
        await llm.ainvoke(messages)


@patch("neo4j_graphrag.llm.mistralai_llm.Mistral")
def test_mistralai_llm_invoke_v2_validation_error(mock_mistral: Mock) -> None:
    """Test V2 interface invoke with invalid message role raises error."""
    mock_mistral_instance = mock_mistral.return_value
    chat_response_mock = MagicMock()
    chat_response_mock.choices = [
        MagicMock(message=MagicMock(content="should not reach here"))
    ]
    mock_mistral_instance.chat.complete.return_value = chat_response_mock

    messages: List[LLMMessage] = [
        {"role": "invalid_role", "content": "This should fail."},  # type: ignore[typeddict-item]
    ]

    llm = MistralAILLM(model_name="mistral-model")

    with pytest.raises(ValueError) as exc_info:
        llm.invoke(messages)
    assert "Unknown role: invalid_role" in str(exc_info.value)


@patch("neo4j_graphrag.llm.mistralai_llm.Mistral")
def test_mistralai_llm_invoke_invalid_input_type(_mock_mistral: Mock) -> None:
    """Test that invalid input type raises appropriate error."""
    llm = MistralAILLM(model_name="mistral-model")

    with pytest.raises(ValueError) as exc_info:
        llm.invoke(123)  # type: ignore
    assert "Invalid input type for invoke method" in str(exc_info.value)


@pytest.mark.asyncio
@patch("neo4j_graphrag.llm.mistralai_llm.Mistral")
async def test_mistralai_llm_ainvoke_invalid_input_type(_mock_mistral: Mock) -> None:
    """Test that invalid input type raises appropriate error for async invoke."""
    llm = MistralAILLM(model_name="mistral-model")

    with pytest.raises(ValueError) as exc_info:
        await llm.ainvoke(123)  # type: ignore
    assert "Invalid input type for ainvoke method" in str(exc_info.value)


@patch("neo4j_graphrag.llm.mistralai_llm.Mistral")
def test_mistralai_llm_get_messages_v2_all_roles(_mock_mistral: Mock) -> None:
    """Test get_messages_v2 method handles all message roles correctly."""
    messages: List[LLMMessage] = [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": "Hello"},
        {"role": "assistant", "content": "Hi there!"},
        {"role": "user", "content": "How are you?"},
    ]

    llm = MistralAILLM(model_name="mistral-model")
    result_messages = llm.get_messages_v2(messages)

    # Verify the correct number of messages are returned
    assert len(result_messages) == 4

    # Verify each message type is correctly converted
    assert result_messages[0].content == "You are a helpful assistant."
    assert result_messages[1].content == "Hello"
    assert result_messages[2].content == "Hi there!"
    assert result_messages[3].content == "How are you?"


@patch("neo4j_graphrag.llm.mistralai_llm.Mistral")
def test_mistralai_llm_get_messages_v2_unknown_role(_mock_mistral: Mock) -> None:
    """Test get_messages_v2 method raises error for unknown role."""
    messages: List[LLMMessage] = [
        {"role": "unknown_role", "content": "This should fail."},  # type: ignore[typeddict-item]
    ]

    llm = MistralAILLM(model_name="mistral-model")

    with pytest.raises(ValueError) as exc_info:
        llm.get_messages_v2(messages)
    assert "Unknown role: unknown_role" in str(exc_info.value)


@patch("neo4j_graphrag.llm.mistralai_llm.Mistral")
def test_mistralai_invoke_v2_with_response_format_raises_error(
    mock_mistral: Mock,
) -> None:
    """Test V2 interface raises NotImplementedError when response_format is used."""

    class TestModel(BaseModel):
        model_config = ConfigDict(extra="forbid")
        value: str

    messages: List[LLMMessage] = [{"role": "user", "content": "Test"}]
    llm = MistralAILLM(api_key="test", model_name="mistral-model")

    with pytest.raises(NotImplementedError) as exc_info:
        llm.invoke(messages, response_format=TestModel)

    assert "MistralAILLM does not currently support structured output" in str(
        exc_info.value
    )


@patch("neo4j_graphrag.llm.mistralai_llm.SDKError", MockSDKError)
@patch("neo4j_graphrag.llm.mistralai_llm.Mistral")
def test_mistralai_invoke_v2_rate_limit_handler_called(
    mock_mistral: Mock,
) -> None:
    """Test that the rate limit handler is invoked on the V2 (List[LLMMessage]) path."""
    messages: List[LLMMessage] = [{"role": "user", "content": "Hello"}]
    mock_mistral_instance = mock_mistral.return_value
    chat_response_mock = MagicMock()
    chat_response_mock.choices = [MagicMock(message=MagicMock(content="Hi there!"))]
    mock_mistral_instance.chat.complete.return_value = chat_response_mock

    spy_handler = MagicMock(wraps=NoOpRateLimitHandler())
    llm = MistralAILLM(model_name="mistral-model", rate_limit_handler=spy_handler)
    response = llm.invoke(messages)

    assert response.content == "Hi there!"
    spy_handler.handle_sync.assert_called_once()


@pytest.mark.asyncio
@patch("neo4j_graphrag.llm.mistralai_llm.SDKError", MockSDKError)
@patch("neo4j_graphrag.llm.mistralai_llm.Mistral")
async def test_mistralai_ainvoke_v2_rate_limit_handler_called(
    mock_mistral: Mock,
) -> None:
    """Test that the rate limit handler is invoked on the async V2 (List[LLMMessage]) path."""
    messages: List[LLMMessage] = [{"role": "user", "content": "Hello"}]
    mock_mistral_instance = mock_mistral.return_value
    chat_response_mock = MagicMock()
    chat_response_mock.choices = [MagicMock(message=MagicMock(content="Hi there!"))]
    mock_mistral_instance.chat.complete_async = AsyncMock(
        return_value=chat_response_mock
    )

    spy_handler = MagicMock(wraps=NoOpRateLimitHandler())
    llm = MistralAILLM(model_name="mistral-model", rate_limit_handler=spy_handler)
    response = await llm.ainvoke(messages)

    assert response.content == "Hi there!"
    spy_handler.handle_async.assert_called_once()



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""
Example demonstrating how to use the ToolsRetriever with a Cypher template.

This example shows:
1. How to create a tool from a Cypher template
2. How to use the ToolsRetriever to select and execute tools based on a query
"""

import os
from typing import Any, Callable, Dict, Optional

import neo4j
from dotenv import load_dotenv

from neo4j_graphrag.llm.openai_llm import OpenAILLM
from neo4j_graphrag.retrievers.base import Retriever
from neo4j_graphrag.retrievers.tools_retriever import ToolsRetriever
from neo4j_graphrag.types import RawSearchResult, RetrieverResultItem

# Load environment variables from .env file (OPENAI_API_KEY required for this example)
load_dotenv()

# Define database credentials
URI = "neo4j+s://demo.neo4jlabs.com"
AUTH = ("recommendations", "recommendations")

# Cypher template to count actors in a specific movie
CYPHER_TEMPLATE = """
MATCH (m:Movie {title: $title})
    OPTIONAL MATCH (m)<-[:ACTED_IN]-(a:Actor)
    WITH m, collect(a.name) AS actor_names, count(a) AS actor_count
    RETURN m.title AS movie_title,
        m.plot AS plot,
        m.released AS year,
        actor_count,
        actor_names
"""


class CypherTemplateRetriever(Retriever):
    """
    Custom retriever that executes a parameterized Cypher query template.
    """

    def __init__(
        self,
        driver: neo4j.Driver,
        cypher_template: str,
        neo4j_database: Optional[str] = None,
        result_formatter: Optional[
            Callable[[neo4j.Record], RetrieverResultItem]
        ] = None,
    ):
        """
        Args:
            driver: Neo4j driver instance
            cypher_template: Cypher query with parameters (e.g., "MATCH (m:Movie {title: $title}) RETURN m")
            neo4j_database: Optional database name
            result_formatter: Optional function to format results
        """
        super().__init__(driver, neo4j_database)
        self.cypher_template = cypher_template
        self.result_formatter = result_formatter

    def get_search_results(
        self,
        query_vector: Optional[list[float]] = None,
        query_text: Optional[str] = None,
        top_k: int = 5,
        query_params: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> RawSearchResult:
        """
        Execute the Cypher template with provided parameters.

        Args:
            query_text: Can be used as a parameter (e.g., for movie title)
            query_params: Dictionary of parameters for the Cypher query
            **kwargs: Additional parameters

        Returns:
            RawSearchResult containing neo4j.Record objects
        """
        # Prepare parameters for the Cypher query
        parameters = query_params or {}

        # Optionally use query_text as a parameter
        if query_text and "query_text" not in parameters:
            parameters["query_text"] = query_text

        # Add any additional kwargs as parameters
        parameters.update(kwargs)

        # Execute the query
        try:
            records, summary, keys = self.driver.execute_query(
                self.cypher_template,
                parameters_=parameters,
                database_=self.neo4j_database,
                routing_=neo4j.RoutingControl.READ,
            )

            return RawSearchResult(
                records=records,
                metadata={
                    "cypher_query": self.cypher_template,
                    "parameters": parameters,
                },
            )
        except Exception as e:
            raise RuntimeError(f"Failed to execute Cypher template: {e}") from e


def main() -> None:
    """Run the example."""
    driver = neo4j.GraphDatabase.driver(URI, auth=AUTH)
    try:
        neo4j_retriever = CypherTemplateRetriever(
            driver=driver,
            cypher_template=CYPHER_TEMPLATE,
        )

        # Convert retriever to tool
        neo4j_tool = neo4j_retriever.convert_to_tool(
            name="movie_info_tool",
            description=(
                "Retrieves the total number of actors in a specific movie and returns "
                "aggregation metrics including actor count, list of actor names, and movie details. "
                "Use this when the user asks about cast size, number of actors, or movie cast information."
            ),
            parameter_descriptions={
                "query_params": (
                    "Dictionary containing 'title' key with the movie title to analyze. "
                    "Example: {'title': 'The Matrix'}"
                )
            },
        )

        llm = OpenAILLM(
            api_key=os.getenv("OPENAI_API_KEY"),
            model_name="gpt-5",
            model_params={"temperature": 0.2},
        )

        tools_retriever = ToolsRetriever(
            driver=driver,
            llm=llm,
            tools=[neo4j_tool],
        )

        query_text = "How many actors are there in Around the World in 80 Days?"
        print("Query:", query_text)
        result = tools_retriever.search(query_text=query_text, return_context=True)
        print("Result:", result)
    finally:
        driver.close()


if __name__ == "__main__":
    main()



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import enum
import json
import logging
from typing import Optional, Any, TypeVar, Generic, Union

from pydantic import validate_call, BaseModel

from neo4j_graphrag.experimental.components.schema import (
    GraphSchema,
    Pattern,
    PropertyType,
    NodeType,
    RelationshipType,
)
from neo4j_graphrag.experimental.components.types import (
    Neo4jGraph,
    Neo4jNode,
    Neo4jRelationship,
    LexicalGraphConfig,
)
from neo4j_graphrag.experimental.pipeline import Component, DataModel

logger = logging.getLogger(__name__)


class PruningReason(str, enum.Enum):
    NOT_IN_SCHEMA = "NOT_IN_SCHEMA"
    MISSING_REQUIRED_PROPERTY = "MISSING_REQUIRED_PROPERTY"
    NO_PROPERTY_LEFT = "NO_PROPERTY_LEFT"
    INVALID_START_OR_END_NODE = "INVALID_START_OR_END_NODE"
    INVALID_PATTERN = "INVALID_PATTERN"
    MISSING_LABEL = "MISSING_LABEL"


ItemType = TypeVar("ItemType")


class PrunedItem(BaseModel, Generic[ItemType]):
    label: str
    item: ItemType
    pruned_reason: PruningReason
    metadata: dict[str, Any] = {}


class PruningStats(BaseModel):
    pruned_nodes: list[PrunedItem[Neo4jNode]] = []
    pruned_relationships: list[PrunedItem[Neo4jRelationship]] = []
    pruned_properties: list[PrunedItem[str]] = []

    @property
    def number_of_pruned_nodes(self) -> int:
        return len(self.pruned_nodes)

    @property
    def number_of_pruned_relationships(self) -> int:
        return len(self.pruned_relationships)

    @property
    def number_of_pruned_properties(self) -> int:
        return len(self.pruned_properties)

    def __str__(self) -> str:
        return (
            f"PruningStats: nodes: {self.number_of_pruned_nodes}, "
            f"relationships: {self.number_of_pruned_relationships}, "
            f"properties: {self.number_of_pruned_properties}"
        )

    def add_pruned_node(
        self, node: Neo4jNode, reason: PruningReason, **kwargs: Any
    ) -> None:
        self.pruned_nodes.append(
            PrunedItem(
                label=node.label, item=node, pruned_reason=reason, metadata=kwargs
            )
        )

    def add_pruned_relationship(
        self, relationship: Neo4jRelationship, reason: PruningReason, **kwargs: Any
    ) -> None:
        self.pruned_relationships.append(
            PrunedItem(
                label=relationship.type,
                item=relationship,
                pruned_reason=reason,
                metadata=kwargs,
            )
        )

    def add_pruned_property(
        self, prop: str, label: str, reason: PruningReason, **kwargs: Any
    ) -> None:
        self.pruned_properties.append(
            PrunedItem(label=label, item=prop, pruned_reason=reason, metadata=kwargs)
        )

    def add_pruned_item(
        self,
        item: Union[Neo4jNode, Neo4jRelationship],
        reason: PruningReason,
        **kwargs: Any,
    ) -> None:
        if isinstance(item, Neo4jNode):
            self.add_pruned_node(
                item,
                reason=reason,
                **kwargs,
            )
        else:
            self.add_pruned_relationship(
                item,
                reason=reason,
                **kwargs,
            )


class GraphPruningResult(DataModel):
    graph: Neo4jGraph
    pruning_stats: PruningStats


class GraphPruning(Component):
    @validate_call
    async def run(
        self,
        graph: Neo4jGraph,
        schema: Optional[GraphSchema] = None,
        lexical_graph_config: Optional[LexicalGraphConfig] = None,
    ) -> GraphPruningResult:
        if lexical_graph_config is None:
            lexical_graph_config = LexicalGraphConfig()
        if schema is not None:
            new_graph, pruning_stats = self._clean_graph(
                graph, schema, lexical_graph_config
            )
        else:
            new_graph = graph
            pruning_stats = PruningStats()
        return GraphPruningResult(
            graph=new_graph,
            pruning_stats=pruning_stats,
        )

    def _clean_graph(
        self,
        graph: Neo4jGraph,
        schema: GraphSchema,
        lexical_graph_config: LexicalGraphConfig,
    ) -> tuple[Neo4jGraph, PruningStats]:
        """
        Verify that the graph conforms to the provided schema.

        Remove invalid entities,relationships, and properties.
        If an entity is removed, all of its relationships are also removed.
        If no valid properties remain for an entity, remove that entity.
        """
        pruning_stats = PruningStats()
        filtered_nodes = self._enforce_nodes(
            graph.nodes,
            schema,
            lexical_graph_config,
            pruning_stats,
        )
        if not filtered_nodes:
            logger.warning(
                "PRUNING: all nodes were pruned, resulting graph will be empty. Check logs for details."
            )
            return Neo4jGraph(), pruning_stats

        filtered_rels = self._enforce_relationships(
            graph.relationships,
            filtered_nodes,
            schema,
            lexical_graph_config,
            pruning_stats,
        )

        return (
            Neo4jGraph(nodes=filtered_nodes, relationships=filtered_rels),
            pruning_stats,
        )

    def _validate_node(
        self,
        node: Neo4jNode,
        pruning_stats: PruningStats,
        schema_entity: Optional[NodeType],
        additional_node_types: bool,
    ) -> Optional[Neo4jNode]:
        if not node.label:
            pruning_stats.add_pruned_node(node, reason=PruningReason.MISSING_LABEL)
            return None
        if not node.id:
            pruning_stats.add_pruned_node(
                node,
                reason=PruningReason.MISSING_REQUIRED_PROPERTY,
                missing_required_properties=["id"],
                details="The node was extracted without a valid ID.",
            )
            return None
        if not schema_entity:
            # node type not declared in the schema
            if additional_node_types:
                # keep node as it is as we do not have any additional info
                return node
            # it's not in schema
            pruning_stats.add_pruned_node(node, reason=PruningReason.NOT_IN_SCHEMA)
            return None
        filtered_props = self._enforce_properties(
            node,
            schema_entity,
            pruning_stats,
            prune_empty=True,
        )
        if not filtered_props:
            return None
        return Neo4jNode(
            id=node.id,
            label=node.label,
            properties=filtered_props,
            embedding_properties=node.embedding_properties,
        )

    def _enforce_nodes(
        self,
        nodes: list[Neo4jNode],
        schema: GraphSchema,
        lexical_graph_config: LexicalGraphConfig,
        pruning_stats: PruningStats,
    ) -> list[Neo4jNode]:
        """
        Filter nodes to be conformant to the schema.

        Keep only those whose label is in schema
        (unless schema has additional_node_types=True, default value)
        For each valid node, validate properties. If a node is left without
        properties, prune it.
        """
        valid_nodes = []
        for node in nodes:
            if node.label in lexical_graph_config.lexical_graph_node_labels:
                valid_nodes.append(node)
                continue
            schema_entity = schema.node_type_from_label(node.label)
            new_node = self._validate_node(
                node,
                pruning_stats,
                schema_entity,
                additional_node_types=schema.additional_node_types,
            )
            if new_node:
                valid_nodes.append(new_node)
        return valid_nodes

    def _validate_relationship(
        self,
        rel: Neo4jRelationship,
        valid_nodes: dict[str, str],
        pruning_stats: PruningStats,
        relationship_type: Optional[RelationshipType],
        additional_relationship_types: bool,
        patterns: tuple[Pattern, ...],
        additional_patterns: bool,
    ) -> Optional[Neo4jRelationship]:
        if not rel.type:
            pruning_stats.add_pruned_relationship(
                rel, reason=PruningReason.MISSING_LABEL
            )
            return None
        # validate start/end node IDs are valid nodes
        if rel.start_node_id not in valid_nodes or rel.end_node_id not in valid_nodes:
            logger.debug(
                f"PRUNING:: {rel} as one of {rel.start_node_id} or {rel.end_node_id} is not a valid node"
            )
            pruning_stats.add_pruned_relationship(
                rel, reason=PruningReason.INVALID_START_OR_END_NODE
            )
            return None

        # validate relationship type
        if relationship_type is None:
            if not additional_relationship_types:
                logger.debug(
                    f"PRUNING:: {rel} as {rel.type} is not in the schema and `additional_relationship_types` is False"
                )
                pruning_stats.add_pruned_relationship(
                    rel, reason=PruningReason.NOT_IN_SCHEMA
                )
                return None

        # validate pattern
        tuple_valid = True
        reverse_tuple_valid = False
        if patterns and relationship_type:
            start_label = valid_nodes[rel.start_node_id]
            end_label = valid_nodes[rel.end_node_id]
            tuple_valid = (start_label, rel.type, end_label) in patterns
            # try to reverse relationship only if initial order is not valid
            reverse_tuple_valid = (
                not tuple_valid
                and (
                    end_label,
                    rel.type,
                    start_label,
                )
                in patterns
            )

        if not tuple_valid and not reverse_tuple_valid and not additional_patterns:
            logger.debug(f"PRUNING:: {rel} not in the allowed patterns")
            pruning_stats.add_pruned_relationship(
                rel, reason=PruningReason.INVALID_PATTERN
            )
            return None

        # filter properties if we can
        if relationship_type is not None:
            filtered_props = self._enforce_properties(
                rel,
                relationship_type,
                pruning_stats,
                prune_empty=False,
            )
        else:
            filtered_props = rel.properties

        return Neo4jRelationship(
            start_node_id=rel.end_node_id if reverse_tuple_valid else rel.start_node_id,
            end_node_id=rel.start_node_id if reverse_tuple_valid else rel.end_node_id,
            type=rel.type,
            properties=filtered_props,
            embedding_properties=rel.embedding_properties,
        )

    def _enforce_relationships(
        self,
        relationships: list[Neo4jRelationship],
        filtered_nodes: list[Neo4jNode],
        schema: GraphSchema,
        lexical_graph_config: LexicalGraphConfig,
        pruning_stats: PruningStats,
    ) -> list[Neo4jRelationship]:
        """
        Filter relationships to be conformant to the schema.

        Keep only those whose types are in schema, start/end node conform to schema,
        and start/end nodes are in filtered nodes (i.e., kept after node enforcement).
        For each valid relationship, filter out properties not present in the schema.

        If a relationship direction is incorrect, invert it.
        """

        valid_rels = []
        valid_nodes = {node.id: node.label for node in filtered_nodes}
        for rel in relationships:
            if rel.type in lexical_graph_config.lexical_graph_relationship_types:
                # Need to validate start/end nodes exist for lexical relationships
                if rel.start_node_id in valid_nodes and rel.end_node_id in valid_nodes:
                    valid_rels.append(rel)
                else:
                    logger.debug(
                        f"PRUNING:: Lexical relationship {rel} as one of {rel.start_node_id} or {rel.end_node_id} is not a valid node"
                    )
                    pruning_stats.add_pruned_relationship(
                        rel, reason=PruningReason.INVALID_START_OR_END_NODE
                    )
                continue
            schema_relation = schema.relationship_type_from_label(rel.type)
            new_rel = self._validate_relationship(
                rel,
                valid_nodes,
                pruning_stats,
                schema_relation,
                schema.additional_relationship_types,
                schema.patterns,
                schema.additional_patterns,
            )
            if new_rel:
                valid_rels.append(new_rel)
        return valid_rels

    def _enforce_properties(
        self,
        item: Union[Neo4jNode, Neo4jRelationship],
        schema_item: Union[NodeType, RelationshipType],
        pruning_stats: PruningStats,
        prune_empty: bool = False,
    ) -> dict[str, Any]:
        """
        Enforce properties:
        - Ensure property type: for now, just prevent having invalid property types (e.g. map)
        - Filter out those that are not in schema (i.e., valid properties) if allowed properties is False.
        - Check that all required properties are present and not null.
        """
        type_safe_properties = self._ensure_property_types(
            item.properties,
        )
        filtered_properties = self._filter_properties(
            type_safe_properties,
            schema_item.properties,
            schema_item.additional_properties,
            item.token,  # label or type
            pruning_stats,
        )
        if not filtered_properties and prune_empty:
            pruning_stats.add_pruned_item(item, reason=PruningReason.NO_PROPERTY_LEFT)
            return filtered_properties
        missing_required_properties = self._check_required_properties(
            filtered_properties,
            valid_properties=schema_item.properties,
        )
        if missing_required_properties:
            pruning_stats.add_pruned_item(
                item,
                reason=PruningReason.MISSING_REQUIRED_PROPERTY,
                missing_required_properties=missing_required_properties,
            )
            return {}
        return filtered_properties

    def _filter_properties(
        self,
        properties: dict[str, Any],
        valid_properties: list[PropertyType],
        additional_properties: bool,
        node_label: str,
        pruning_stats: PruningStats,
    ) -> dict[str, Any]:
        """Filters out properties not in schema if additional_properties is False"""
        if additional_properties:
            # we do not need to filter any property, just return the initial properties
            return properties
        valid_prop_names = {prop.name for prop in valid_properties}
        filtered_properties = {}
        for prop_name, prop_value in properties.items():
            if prop_name not in valid_prop_names:
                pruning_stats.add_pruned_property(
                    prop_name,
                    node_label,
                    reason=PruningReason.NOT_IN_SCHEMA,
                    value=prop_value,
                )
                continue
            filtered_properties[prop_name] = prop_value
        return filtered_properties

    def _check_required_properties(
        self, filtered_properties: dict[str, Any], valid_properties: list[PropertyType]
    ) -> list[str]:
        """Returns the list of missing required properties, if any."""
        required_prop_names = {prop.name for prop in valid_properties if prop.required}
        missing_required_properties = []
        for req_prop in required_prop_names:
            if filtered_properties.get(req_prop) is None:
                missing_required_properties.append(req_prop)
        return missing_required_properties

    def _ensure_property_types(
        self,
        filtered_properties: dict[str, Any],
    ) -> dict[str, Any]:
        type_safe_properties = {}
        for prop_name, prop_value in filtered_properties.items():
            if isinstance(prop_value, dict):
                # just ensure the type will not raise error on insert, while preserving data
                type_safe_properties[prop_name] = json.dumps(prop_value, default=str)
                continue

            # this is where we could check types of other properties
            # but keep it simple for now
            type_safe_properties[prop_name] = prop_value
        return type_safe_properties



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Any

import pytest

from neo4j_graphrag.experimental.components.graph_pruning import GraphPruning
from neo4j_graphrag.experimental.components.schema import GraphSchema
from neo4j_graphrag.experimental.components.types import (
    Neo4jGraph,
    Neo4jNode,
    Neo4jRelationship,
)


@pytest.fixture
def extracted_graph() -> Neo4jGraph:
    """This is the graph to be pruned in all the below tests,
    using different schema configuration.
    """
    return Neo4jGraph(
        nodes=[
            Neo4jNode(
                id="1",
                label="Person",
                properties={
                    "name": "John Doe",
                },
            ),
            Neo4jNode(
                id="2",
                label="Person",
                properties={
                    "height": 180,
                },
            ),
            Neo4jNode(
                id="3",
                label="Person",
                properties={
                    "name": "Jane Doe",
                    "weight": 90,
                },
            ),
            Neo4jNode(
                id="10",
                label="Organization",
                properties={
                    "name": "Azerty Inc.",
                    "created": 1999,
                },
            ),
        ],
        relationships=[
            Neo4jRelationship(
                start_node_id="1",
                end_node_id="2",
                type="KNOWS",
                properties={"firstMetIn": 2025},
            ),
            Neo4jRelationship(
                start_node_id="1",
                end_node_id="3",
                type="KNOWS",
            ),
            Neo4jRelationship(
                start_node_id="1",
                end_node_id="2",
                type="MANAGES",
            ),
            Neo4jRelationship(
                start_node_id="1",
                end_node_id="10",
                type="MANAGES",
            ),
            Neo4jRelationship(
                start_node_id="1",
                end_node_id="10",
                type="WORKS_FOR",
            ),
        ],
    )


async def _test(
    extracted_graph: Neo4jGraph, schema_dict: dict[str, Any], expected_graph: Neo4jGraph
) -> None:
    schema = GraphSchema.model_validate(schema_dict)
    pruner = GraphPruning()
    res = await pruner.run(extracted_graph, schema)
    assert res.graph == expected_graph


@pytest.mark.asyncio
async def test_graph_pruning_loose(extracted_graph: Neo4jGraph) -> None:
    """Loose schema:
    - no required properties
    - all additional* allowed

    => we keep everything from the extracted graph
    """
    schema_dict = {
        "node_types": [
            {
                "label": "Person",
                "properties": [
                    {"name": "name", "type": "STRING"},
                    {"name": "height", "type": "INTEGER"},
                ],
                "additional_properties": True,
            }
        ],
        "relationship_types": [
            {
                "label": "KNOWS",
            }
        ],
        "patterns": [
            ("Person", "KNOWS", "Person"),
        ],
        "additional_node_types": True,
        "additional_relationship_types": True,
        "additional_patterns": True,
    }
    await _test(extracted_graph, schema_dict, extracted_graph)


@pytest.mark.asyncio
async def test_graph_pruning_missing_required_property(
    extracted_graph: Neo4jGraph,
) -> None:
    """Person node type has a required 'name' property:
    - extracted nodes without this property are pruned
    - any relationship tied to this node is also pruned
    """
    schema_dict = {
        "node_types": [
            {
                "label": "Person",
                "properties": [
                    {
                        "name": "name",
                        "type": "STRING",
                        "required": True,
                    },
                    {"name": "height", "type": "INTEGER"},
                ],
                "additional_properties": True,
            }
        ],
        "relationship_types": [
            {
                "label": "KNOWS",
            }
        ],
        "patterns": [
            ("Person", "KNOWS", "Person"),
        ],
        "additional_node_types": True,
        "additional_relationship_types": True,
        "additional_patterns": True,
    }
    filtered_graph = Neo4jGraph(
        nodes=[
            Neo4jNode(
                id="1",
                label="Person",
                properties={
                    "name": "John Doe",
                },
            ),
            # do not have the required "name" property
            # Neo4jNode(
            #     id="2",
            #     label="Person",
            #     properties={
            #         "height": 180,
            #     }
            # ),
            Neo4jNode(
                id="3",
                label="Person",
                properties={
                    "name": "Jane Doe",
                    "weight": 90,
                },
            ),
            Neo4jNode(
                id="10",
                label="Organization",
                properties={
                    "name": "Azerty Inc.",
                    "created": 1999,
                },
            ),
        ],
        relationships=[
            # node "2" was pruned
            # Neo4jRelationship(
            #     start_node_id="1",
            #     end_node_id="2",
            #     type="KNOWS",
            #     properties={"firstMetIn": 2025},
            # ),
            Neo4jRelationship(
                start_node_id="1",
                end_node_id="3",
                type="KNOWS",
            ),
            # node "2" was pruned
            # Neo4jRelationship(
            #     start_node_id="1",
            #     end_node_id="2",
            #     type="MANAGES",
            # ),
            Neo4jRelationship(
                start_node_id="1",
                end_node_id="10",
                type="MANAGES",
            ),
            Neo4jRelationship(
                start_node_id="1",
                end_node_id="10",
                type="WORKS_FOR",
            ),
        ],
    )
    await _test(extracted_graph, schema_dict, filtered_graph)


@pytest.mark.asyncio
async def test_graph_pruning_strict_properties_and_node_types(
    extracted_graph: Neo4jGraph,
) -> None:
    """Additional properties on Person nodes are not allowed.
    Additional node types are not allowed.

    => we prune "Organization" nodes (not in schema)
    and the "weight" property that was extracted for some persons.
    """
    schema_dict = {
        "node_types": [
            {
                "label": "Person",
                "properties": [
                    {
                        "name": "name",
                        "type": "STRING",
                    },
                    {"name": "height", "type": "INTEGER"},
                ],
                # "additional_properties": False,  # default value
            }
        ],
        "relationship_types": [
            {
                "label": "KNOWS",
            }
        ],
        "patterns": [
            ("Person", "KNOWS", "Person"),
        ],
        # "additional_node_types": False,  # default value
        "additional_relationship_types": True,
        "additional_patterns": True,
    }
    filtered_graph = Neo4jGraph(
        nodes=[
            Neo4jNode(
                id="1",
                label="Person",
                properties={
                    "name": "John Doe",
                },
            ),
            Neo4jNode(
                id="2",
                label="Person",
                properties={
                    "height": 180,
                },
            ),
            Neo4jNode(
                id="3",
                label="Person",
                properties={
                    "name": "Jane Doe",
                    # weight not in listed properties
                    # "weight": 90,
                },
            ),
            # label "Organization" not in schema
            # Neo4jNode(
            #     id="10",
            #     label="Organization",
            #     properties={
            #         "name": "Azerty Inc.",
            #         "created": 1999,
            #     }
            # ),
        ],
        relationships=[
            Neo4jRelationship(
                start_node_id="1",
                end_node_id="2",
                type="KNOWS",
                properties={"firstMetIn": 2025},
            ),
            Neo4jRelationship(
                start_node_id="1",
                end_node_id="3",
                type="KNOWS",
            ),
            Neo4jRelationship(
                start_node_id="1",
                end_node_id="2",
                type="MANAGES",
            ),
            # node "10" was pruned (label not allowed)
            # Neo4jRelationship(
            #     start_node_id="1",
            #     end_node_id="10",
            #     type="MANAGES",
            # ),
            # Neo4jRelationship(
            #     start_node_id="1",
            #     end_node_id="10",
            #     type="WORKS_FOR",
            # )
        ],
    )
    await _test(extracted_graph, schema_dict, filtered_graph)


@pytest.mark.asyncio
async def test_graph_pruning_strict_patterns(extracted_graph: Neo4jGraph) -> None:
    """Additional patterns not allowed:

    - MANAGES: it's a known relationship type but without any pattern, it's pruned
    - WORKS_FOR: it's not a known relationship type, and additional_relationship_types is allowed
        so we keep it.
    """
    # - no additional patterns allowed
    schema_dict = {
        "node_types": [
            {
                "label": "Person",
                "properties": [
                    {
                        "name": "name",
                        "type": "STRING",
                    },
                    {"name": "height", "type": "INTEGER"},
                ],
                "additional_properties": True,
            },
            {
                "label": "Organization",
            },
        ],
        "relationship_types": [
            {
                "label": "KNOWS",
            },
            {
                "label": "MANAGES",
            },
        ],
        "patterns": (
            ("Person", "KNOWS", "Person"),
            ("Person", "KNOWS", "Organization"),
        ),
        "additional_node_types": True,
        "additional_relationship_types": False,
        "additional_patterns": False,
    }
    filtered_graph = Neo4jGraph(
        nodes=[
            Neo4jNode(
                id="1",
                label="Person",
                properties={
                    "name": "John Doe",
                },
            ),
            Neo4jNode(
                id="2",
                label="Person",
                properties={
                    "height": 180,
                },
            ),
            Neo4jNode(
                id="3",
                label="Person",
                properties={
                    "name": "Jane Doe",
                    "weight": 90,
                },
            ),
            Neo4jNode(
                id="10",
                label="Organization",
                properties={
                    "name": "Azerty Inc.",
                    "created": 1999,
                },
            ),
        ],
        relationships=[
            Neo4jRelationship(
                start_node_id="1",
                end_node_id="2",
                type="KNOWS",
                properties={"firstMetIn": 2025},
            ),
            Neo4jRelationship(
                start_node_id="1",
                end_node_id="3",
                type="KNOWS",
            ),
            # invalid pattern (person, manages, person)
            # Neo4jRelationship(
            #     start_node_id="1",
            #     end_node_id="2",
            #     type="MANAGES",
            # ),
            # invalid pattern (person, works for, person)
            # Neo4jRelationship(
            #     start_node_id="1",
            #     end_node_id="10",
            #     type="WORKS_FOR",
            # ),
        ],
    )
    await _test(extracted_graph, schema_dict, filtered_graph)



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

import datetime
from typing import Any, Optional
from unittest.mock import ANY, Mock, patch

import pytest
from neo4j_graphrag.experimental.components.graph_pruning import (
    GraphPruning,
    GraphPruningResult,
    PruningStats,
)
from neo4j_graphrag.experimental.components.schema import (
    GraphSchema,
    NodeType,
    Pattern,
    PropertyType,
    RelationshipType,
)
from neo4j_graphrag.experimental.components.types import (
    LexicalGraphConfig,
    Neo4jGraph,
    Neo4jNode,
    Neo4jRelationship,
)


@pytest.fixture(scope="module")
def lexical_graph_config() -> LexicalGraphConfig:
    return LexicalGraphConfig(
        chunk_node_label="Paragraph",
    )


@pytest.mark.parametrize(
    "properties, valid_properties, additional_properties, expected_filtered_properties",
    [
        (
            # no required, additional allowed
            {
                "name": "John Does",
                "age": 25,
            },
            [
                PropertyType(
                    name="name",
                    type="STRING",
                )
            ],
            True,
            {
                "name": "John Does",
                "age": 25,
            },
        ),
        (
            # no required, additional not allowed
            {
                "name": "John Does",
                "age": 25,
            },
            [
                PropertyType(
                    name="name",
                    type="STRING",
                )
            ],
            False,
            {
                "name": "John Does",
            },
        ),
    ],
)
def test_graph_pruning_filter_properties(
    properties: dict[str, Any],
    valid_properties: list[PropertyType],
    additional_properties: bool,
    expected_filtered_properties: dict[str, Any],
) -> None:
    pruner = GraphPruning()
    filtered_properties = pruner._filter_properties(
        properties,
        valid_properties,
        additional_properties=additional_properties,
        node_label="Label",
        pruning_stats=PruningStats(),
    )
    assert filtered_properties == expected_filtered_properties


@pytest.mark.parametrize(
    "properties, expected_filtered_properties",
    [
        (
            # all good, no bad types
            {
                "name": "John Does",
                "age": 25,
                "is_active": True,
            },
            {
                "name": "John Does",
                "age": 25,
                "is_active": True,
            },
        ),
        (
            # map must be serialized
            {
                "age": {"dob": datetime.date(2000, 1, 1), "age_in_2025": 25},
            },
            {
                "age": '{"dob": "2000-01-01", "age_in_2025": 25}',
            },
        ),
    ],
)
def test_graph_pruning_ensure_property_type(
    properties: dict[str, Any],
    expected_filtered_properties: dict[str, Any],
) -> None:
    pruner = GraphPruning()
    type_safe_properties = pruner._ensure_property_types(
        properties,
    )
    assert type_safe_properties == expected_filtered_properties


@pytest.fixture(scope="module")
def node_type_no_properties() -> NodeType:
    return NodeType(label="Person")


@pytest.fixture(scope="module")
def node_type_required_name() -> NodeType:
    return NodeType(
        label="Person",
        properties=[
            PropertyType(name="name", type="STRING", required=True),
            PropertyType(name="age", type="INTEGER"),
        ],
    )


@pytest.mark.parametrize(
    "node, entity, additional_node_types, expected_node",
    [
        # all good, with default values
        (
            Neo4jNode(id="1", label="Person", properties={"name": "John Doe"}),
            "node_type_no_properties",
            True,
            Neo4jNode(id="1", label="Person", properties={"name": "John Doe"}),
        ),
        # properties empty (missing default)
        (
            Neo4jNode(id="1", label="Person", properties={"age": 45}),
            "node_type_required_name",
            True,
            None,
        ),
        # node label not is schema, additional not allowed
        (
            Neo4jNode(id="1", label="Location", properties={"name": "New York"}),
            None,
            False,
            None,
        ),
        # node label not is schema, additional allowed
        (
            Neo4jNode(id="1", label="Location", properties={"name": "New York"}),
            None,
            True,
            Neo4jNode(id="1", label="Location", properties={"name": "New York"}),
        ),
        # node label not valid
        (
            Neo4jNode(id="1", label="", properties={"name": "New York"}),
            "node_type_required_name",
            True,
            None,
        ),
        # node ID not valid
        (
            Neo4jNode(id="", label="Location", properties={"name": "New York"}),
            "node_type_required_name",
            True,
            None,
        ),
    ],
)
def test_graph_pruning_validate_node(
    node: Neo4jNode,
    entity: str,
    additional_node_types: bool,
    expected_node: Neo4jNode,
    request: pytest.FixtureRequest,
) -> None:
    e = request.getfixturevalue(entity) if entity else None

    pruner = GraphPruning()
    result = pruner._validate_node(node, PruningStats(), e, additional_node_types)
    if expected_node is not None:
        assert result == expected_node
    else:
        assert result is None


def test_graph_pruning_enforce_nodes_lexical_graph(
    lexical_graph_config: LexicalGraphConfig,
) -> None:
    pruner = GraphPruning()
    result = pruner._enforce_nodes(
        nodes=[
            Neo4jNode(id="1", label="Paragraph"),
        ],
        schema=GraphSchema(node_types=tuple(), additional_node_types=False),
        lexical_graph_config=lexical_graph_config,
        pruning_stats=PruningStats(),
    )
    assert len(result) == 1
    assert result[0].label == "Paragraph"


def test_graph_pruning_enforce_relationships_lexical_graph_with_pruned_nodes(
    lexical_graph_config: LexicalGraphConfig,
) -> None:
    """Test that lexical relationships are pruned when their nodes are pruned."""
    pruner = GraphPruning()

    # Create nodes: Chunk (lexical) and Person (extracted, will be pruned)
    nodes = [
        Neo4jNode(id="chunk-1", label="Paragraph", properties={"text": "test"}),
        Neo4jNode(
            id="person-1", label="Person", properties={"age": 30}
        ),  # Missing required 'name'
    ]

    # Person node type requires 'name' property
    person_type = NodeType(
        label="Person",
        properties=[
            PropertyType(name="name", type="STRING", required=True),
            PropertyType(name="age", type="INTEGER"),
        ],
    )
    schema = GraphSchema(node_types=(person_type,))

    # Filter nodes - Person should be pruned due to missing required property
    pruning_stats = PruningStats()
    filtered_nodes = pruner._enforce_nodes(
        nodes=nodes,
        schema=schema,
        lexical_graph_config=lexical_graph_config,
        pruning_stats=pruning_stats,
    )

    # Only Chunk should remain
    assert len(filtered_nodes) == 1
    assert filtered_nodes[0].id == "chunk-1"
    assert pruning_stats.number_of_pruned_nodes == 1

    # Create relationships including FROM_CHUNK to the pruned Person node
    relationships = [
        Neo4jRelationship(
            start_node_id="person-1",
            end_node_id="chunk-1",
            type="FROM_CHUNK",  # Lexical relationship
        ),
        Neo4jRelationship(
            start_node_id="chunk-1",
            end_node_id="person-1",
            type="NEXT_CHUNK",  # Another lexical relationship
        ),
    ]

    # Filter relationships
    pruning_stats_rels = PruningStats()
    filtered_rels = pruner._enforce_relationships(
        relationships=relationships,
        filtered_nodes=filtered_nodes,
        schema=schema,
        lexical_graph_config=lexical_graph_config,
        pruning_stats=pruning_stats_rels,
    )

    # Both lexical relationships should be pruned because person-1 node is missing
    assert len(filtered_rels) == 0
    assert pruning_stats_rels.number_of_pruned_relationships == 2


@pytest.fixture
def neo4j_relationship() -> Neo4jRelationship:
    return Neo4jRelationship(
        start_node_id="1",
        end_node_id="2",
        type="REL",
        properties={},
    )


@pytest.fixture
def neo4j_relationship_invalid_type() -> Neo4jRelationship:
    return Neo4jRelationship(
        start_node_id="1",
        end_node_id="2",
        type="",
        properties={},
    )


@pytest.fixture
def neo4j_reversed_relationship(
    neo4j_relationship: Neo4jRelationship,
) -> Neo4jRelationship:
    return Neo4jRelationship(
        start_node_id=neo4j_relationship.end_node_id,
        end_node_id=neo4j_relationship.start_node_id,
        type=neo4j_relationship.type,
        properties=neo4j_relationship.properties,
    )


@pytest.mark.parametrize(
    "relationship, valid_nodes, relationship_type, additional_relationship_types, patterns, additional_patterns, expected_relationship",
    [
        # all good
        (
            "neo4j_relationship",  # relationship,
            {  # valid_nodes
                "1": "Person",
                "2": "Location",
            },
            RelationshipType(  # relationship_type
                label="REL",
            ),
            True,  # additional_relationship_types
            (
                Pattern(source="Person", relationship="REL", target="Location"),
            ),  # patterns
            True,  # additional_patterns
            "neo4j_relationship",  # expected_relationship
        ),
        # reverse relationship
        (
            "neo4j_reversed_relationship",
            {
                "1": "Person",
                "2": "Location",
            },
            RelationshipType(
                label="REL",
            ),
            True,  # additional_relationship_types
            (Pattern(source="Person", relationship="REL", target="Location"),),
            True,  # additional_patterns
            "neo4j_relationship",
        ),
        # invalid start node ID
        (
            "neo4j_reversed_relationship",
            {
                "10": "Person",
                "2": "Location",
            },
            RelationshipType(
                label="REL",
            ),
            True,  # additional_relationship_types
            (Pattern(source="Person", relationship="REL", target="Location"),),
            True,  # additional_patterns
            None,
        ),
        # invalid type, addition allowed
        (
            "neo4j_relationship",
            {
                "1": "Person",
                "2": "Location",
            },
            None,  # relationship_type
            True,  # additional_relationship_types
            (Pattern(source="Person", relationship="REL", target="Location"),),
            True,  # additional_patterns
            "neo4j_relationship",
        ),
        # invalid type, addition allowed but invalid node ID
        (
            "neo4j_relationship",
            {
                "1": "Person",
            },
            None,  # relationship_type
            True,  # additional_relationship_types
            (Pattern(source="Person", relationship="REL", target="Location"),),
            True,  # additional_patterns
            None,
        ),
        # invalid type, addition not allowed
        (
            "neo4j_relationship",
            {
                "1": "Person",
                "2": "Location",
            },
            None,  # relationship_type
            False,  # additional_relationship_types
            (Pattern(source="Person", relationship="REL", target="Location"),),
            True,  # additional_patterns
            None,
        ),
        # invalid pattern, addition allowed
        (
            "neo4j_relationship",
            {
                "1": "Person",
                "2": "Location",
            },
            RelationshipType(
                label="REL",
            ),
            True,  # additional_relationship_types
            (Pattern(source="Person", relationship="REL", target="Person"),),
            True,  # additional_patterns
            "neo4j_relationship",
        ),
        # invalid pattern, addition not allowed
        (
            "neo4j_relationship",
            {
                "1": "Person",
                "2": "Location",
            },
            RelationshipType(
                label="REL",
            ),
            True,  # additional_relationship_types
            (Pattern(source="Person", relationship="REL", target="Person"),),
            False,  # additional_patterns
            None,
        ),
        # invalid extracted type
        (
            "neo4j_relationship_invalid_type",  # relationship,
            {  # valid_nodes
                "1": "Person",
                "2": "Location",
            },
            RelationshipType(  # relationship_type
                label="REL",
            ),
            True,  # additional_relationship_types
            (
                Pattern(source="Person", relationship="REL", target="Location"),
            ),  # patterns
            True,  # additional_patterns
            None,  # expected_relationship
        ),
    ],
)
def test_graph_pruning_validate_relationship(
    relationship: str,
    valid_nodes: dict[str, str],
    relationship_type: RelationshipType,
    additional_relationship_types: bool,
    patterns: tuple[Pattern, ...],
    additional_patterns: bool,
    expected_relationship: Optional[str],
    request: pytest.FixtureRequest,
) -> None:
    relationship_obj = request.getfixturevalue(relationship)
    expected_relationship_obj = (
        request.getfixturevalue(expected_relationship)
        if expected_relationship
        else None
    )

    pruner = GraphPruning()
    assert (
        pruner._validate_relationship(
            relationship_obj,
            valid_nodes,
            PruningStats(),
            relationship_type,
            additional_relationship_types,
            patterns,
            additional_patterns,
        )
        == expected_relationship_obj
    )


@patch("neo4j_graphrag.experimental.components.graph_pruning.GraphPruning._clean_graph")
@pytest.mark.asyncio
async def test_graph_pruning_run_happy_path(
    mock_clean_graph: Mock,
    node_type_required_name: NodeType,
    lexical_graph_config: LexicalGraphConfig,
) -> None:
    initial_graph = Neo4jGraph(
        nodes=[Neo4jNode(id="1", label="Person"), Neo4jNode(id="2", label="Location")],
    )
    schema = GraphSchema(node_types=(node_type_required_name,))
    cleaned_graph = Neo4jGraph(nodes=[Neo4jNode(id="1", label="Person")])
    mock_clean_graph.return_value = (cleaned_graph, PruningStats())
    pruner = GraphPruning()
    pruner_result = await pruner.run(
        graph=initial_graph,
        schema=schema,
        lexical_graph_config=lexical_graph_config,
    )
    assert isinstance(pruner_result, GraphPruningResult)
    assert pruner_result.graph == cleaned_graph
    mock_clean_graph.assert_called_once_with(
        initial_graph, schema, lexical_graph_config
    )


@pytest.mark.asyncio
async def test_graph_pruning_run_no_schema() -> None:
    initial_graph = Neo4jGraph(nodes=[Neo4jNode(id="1", label="Person")])
    pruner = GraphPruning()
    pruner_result = await pruner.run(
        graph=initial_graph,
        schema=None,
    )
    assert isinstance(pruner_result, GraphPruningResult)
    assert pruner_result.graph == initial_graph


@patch(
    "neo4j_graphrag.experimental.components.graph_pruning.GraphPruning._enforce_nodes"
)
def test_graph_pruning_clean_graph(
    mock_enforce_nodes: Mock,
    lexical_graph_config: LexicalGraphConfig,
) -> None:
    mock_enforce_nodes.return_value = []
    initial_graph = Neo4jGraph(nodes=[Neo4jNode(id="1", label="Person")])
    schema = GraphSchema(node_types=())
    pruner = GraphPruning()
    cleaned_graph, pruning_stats = pruner._clean_graph(
        initial_graph, schema, lexical_graph_config
    )
    assert cleaned_graph == Neo4jGraph()
    assert isinstance(pruning_stats, PruningStats)
    mock_enforce_nodes.assert_called_once_with(
        [Neo4jNode(id="1", label="Person")],
        schema,
        lexical_graph_config,
        ANY,
    )



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from unittest.mock import AsyncMock

import pytest

from neo4j_graphrag.experimental.pipeline import Component
from neo4j_graphrag.experimental.pipeline.types.context import RunContext
from .components import ComponentMultiply, ComponentMultiplyWithContext, IntResultModel


def test_component_inputs() -> None:
    inputs = ComponentMultiply.component_inputs
    assert "number1" in inputs
    assert inputs["number1"]["has_default"] is False
    assert "number2" in inputs
    assert inputs["number2"]["has_default"] is True


def test_component_outputs() -> None:
    outputs = ComponentMultiply.component_outputs
    assert "result" in outputs
    assert outputs["result"]["has_default"] is True
    assert outputs["result"]["annotation"] == int


@pytest.mark.asyncio
async def test_component_run() -> None:
    c = ComponentMultiply()
    result = await c.run(number1=1, number2=2)
    assert isinstance(result, IntResultModel)
    assert isinstance(
        result.result,
        # we know this is a type and not a bool or str:
        ComponentMultiply.component_outputs["result"]["annotation"],  # type: ignore
    )


@pytest.mark.asyncio
async def test_component_run_with_context_default_implementation() -> None:
    c = ComponentMultiply()
    result = await c.run_with_context(
        # context can not be null in the function signature,
        # but it's ignored in this case
        None,  # type: ignore
        number1=1,
        number2=2,
    )
    # the type checker doesn't know about the type
    # because the method is not re-declared
    assert result.result == 2  # type: ignore


@pytest.mark.asyncio
async def test_component_run_with_context() -> None:
    c = ComponentMultiplyWithContext()
    notifier_mock = AsyncMock()
    result = await c.run_with_context(
        RunContext(run_id="run_id", task_name="task_name", notifier=notifier_mock),
        number1=1,
        number2=2,
    )
    assert result.result == 2
    notifier_mock.assert_awaited_once()


def test_component_missing_method() -> None:
    with pytest.raises(RuntimeError) as e:

        class WrongComponent(Component):
            # we must have either run or run_with_context
            pass

    assert (
        "You must implement either `run` or `run_with_context` in Component 'WrongComponent'"
        in str(e)
    )



from neo4j_graphrag.experimental.components.lexical_graph import (
    LexicalGraphBuilder,
)
from neo4j_graphrag.experimental.components.types import (
    GraphResult,
    LexicalGraphConfig,
    TextChunk,
    TextChunks,
)


async def main() -> GraphResult:
    # optionally, define a LexicalGraphConfig object
    # shown below with default values
    config = LexicalGraphConfig(
        chunk_node_label="Chunk",
        document_node_label="Document",
        chunk_to_document_relationship_type="PART_OF_DOCUMENT",
        next_chunk_relationship_type="NEXT_CHUNK",
        node_to_chunk_relationship_type="PART_OF_CHUNK",
        chunk_embedding_property="embeddings",
    )
    builder = LexicalGraphBuilder(
        config=config,  # optional
    )
    graph_result = await builder.run(
        text_chunks=TextChunks(chunks=[TextChunk(text="....", index=0)]),
        # document_info={"path": "example"},  # uncomment to create a "Document" node
    )
    return graph_result



"""End to end example of building a RAG pipeline backed by a Neo4j database.
Requires OPENAI_API_KEY to be in the env var.

This example illustrates:
- VectorCypherRetriever with a custom formatter function to extract relevant
    context from neo4j result
- Logging configuration
"""

import logging

import neo4j
from neo4j_graphrag.embeddings.openai import OpenAIEmbeddings
from neo4j_graphrag.generation import GraphRAG
from neo4j_graphrag.llm import OpenAILLM
from neo4j_graphrag.retrievers import VectorCypherRetriever
from neo4j_graphrag.types import RetrieverResultItem

# Define database credentials
URI = "neo4j+s://demo.neo4jlabs.com"
AUTH = ("recommendations", "recommendations")
DATABASE = "recommendations"
INDEX = "moviePlotsEmbedding"


# setup logger config
logger = logging.getLogger("neo4j_graphrag")
logging.basicConfig(format="%(asctime)s - %(message)s")
logger.setLevel(logging.DEBUG)


def formatter(record: neo4j.Record) -> RetrieverResultItem:
    return RetrieverResultItem(content=f"{record.get('title')}: {record.get('plot')}")


driver = neo4j.GraphDatabase.driver(
    URI,
    auth=AUTH,
)

embedder = OpenAIEmbeddings()

retriever = VectorCypherRetriever(
    driver,
    index_name=INDEX,
    retrieval_query="with node, score return node.title as title, node.plot as plot",
    result_formatter=formatter,
    embedder=embedder,
    neo4j_database=DATABASE,
)

llm = OpenAILLM(model_name="gpt-5", model_params={"temperature": 0})

rag = GraphRAG(retriever=retriever, llm=llm)

result = rag.search(
    "Tell me more about Avatar movies",
    return_context=True,
    # optional
    response_fallback="I can't answer this question without context",
)
print(result.answer)
# print(result.retriever_result)

driver.close()



"""This example demonstrate how to embed a text into a vector
using Google models and the VertexAI API.
"""

from neo4j_graphrag.embeddings import VertexAIEmbeddings

embeder = VertexAIEmbeddings(model="text-embedding-005")
res = embeder.embed_query("my question")
print(res[:10])



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

from typing import Any
from unittest.mock import MagicMock

import neo4j
import pytest
from neo4j import Driver
from neo4j_graphrag.experimental.components.text_splitters.fixed_size_splitter import (
    FixedSizeSplitter,
)
from neo4j_graphrag.experimental.components.types import LexicalGraphConfig
from neo4j_graphrag.experimental.pipeline.kg_builder import SimpleKGPipeline
from neo4j_graphrag.llm import LLMResponse


@pytest.fixture(scope="function", autouse=True)
def clear_db(driver: Driver) -> Any:
    driver.execute_query("MATCH (n) DETACH DELETE n")
    yield


@pytest.fixture(scope="module")
def llm_json_response_3_nodes_2_relationships() -> str:
    return """{
                        "nodes": [
                            {
                                "id": "0",
                                "label": "Person",
                                "properties": {
                                    "name": "Harry Potter"
                                }
                            },
                            {
                                "id": "1",
                                "label": "Person",
                                "properties": {
                                    "name": "Alastor Mad-Eye Moody"
                                }
                            },
                            {
                                "id": "2",
                                "label": "Organization",
                                "properties": {
                                    "name": "The Order of the Phoenix"
                                }
                            }
                        ],
                        "relationships": [
                            {
                                "type": "KNOWS",
                                "start_node_id": "0",
                                "end_node_id": "1"
                            },
                            {
                                "type": "LED_BY",
                                "start_node_id": "2",
                                "end_node_id": "1"
                            }
                        ]
                    }"""


@pytest.mark.asyncio
@pytest.mark.usefixtures("setup_neo4j_for_kg_construction")
async def test_pipeline_builder_happy_path_legacy_schema(
    harry_potter_text: str,
    llm: MagicMock,
    embedder: MagicMock,
    driver: neo4j.Driver,
    llm_json_response_3_nodes_2_relationships: str,
) -> None:
    """When everything works as expected, extracted entities, relations and text
    chunks must be in the DB
    """
    driver.execute_query("MATCH (n) DETACH DELETE n")
    embedder.async_embed_query.return_value = [1, 2, 3]
    llm.ainvoke.side_effect = [
        LLMResponse(
            content=llm_json_response_3_nodes_2_relationships,
        )
    ]

    # Instantiate Entity and Relation objects
    entities = ["Person", "Organization", "Horcrux", "Location"]
    relations = ["SITUATED_AT", "INTERACTS", "OWNS", "LED_BY"]
    potential_schema = [
        ("Person", "SITUATED_AT", "Location"),
        ("Person", "INTERACTS", "Person"),
        ("Person", "OWNS", "Horcrux"),
        ("Organization", "LED_BY", "Person"),
    ]

    # Additional arguments
    lexical_graph_config = LexicalGraphConfig(chunk_node_label="chunkNodeLabel")
    from_pdf = False
    on_error = "RAISE"

    # Create an instance of the SimpleKGPipeline
    kg_builder_text = SimpleKGPipeline(
        llm=llm,
        driver=driver,
        embedder=embedder,
        entities=entities,
        relations=relations,
        potential_schema=potential_schema,
        from_pdf=from_pdf,
        on_error=on_error,
        lexical_graph_config=lexical_graph_config,
    )

    # Run the knowledge graph building process with text input
    await kg_builder_text.run_async(text=harry_potter_text)

    # check the content of the graph:
    # check lexical graph content
    records, _, _ = driver.execute_query("MATCH (start:chunkNodeLabel) RETURN start")
    assert len(records) == 1

    # check entity -> chunk relationships
    records, _, _ = driver.execute_query(
        "MATCH (chunk:chunkNodeLabel)<-[rel:FROM_CHUNK]-(entity:__Entity__) RETURN chunk, rel, entity"
    )
    assert len(records) == 3  # three entities according to mocked LLMResponse


@pytest.mark.asyncio
@pytest.mark.usefixtures("setup_neo4j_for_kg_construction")
async def test_pipeline_builder_happy_path(
    harry_potter_text: str,
    llm: MagicMock,
    embedder: MagicMock,
    driver: neo4j.Driver,
    llm_json_response_3_nodes_2_relationships: str,
) -> None:
    """When everything works as expected, extracted entities, relations and text
    chunks must be in the DB
    """
    driver.execute_query("MATCH (n) DETACH DELETE n")
    embedder.async_embed_query.return_value = [1, 2, 3]
    llm.ainvoke.side_effect = [
        LLMResponse(
            content=llm_json_response_3_nodes_2_relationships,
        )
    ]

    # Instantiate schema
    entities = ["Person"]
    relations: list[str] = []
    potential_schema: list[tuple[str, str, str]] = []
    schema = {
        "node_types": entities,
        "relationship_types": relations,
        "patterns": potential_schema,
        "additional_node_types": False,
    }

    # Additional arguments
    lexical_graph_config = LexicalGraphConfig(chunk_node_label="chunkNodeLabel")
    from_pdf = False
    on_error = "RAISE"

    # Create an instance of the SimpleKGPipeline
    kg_builder_text = SimpleKGPipeline(
        llm=llm,
        driver=driver,
        embedder=embedder,
        schema=schema,  # type: ignore[arg-type]
        from_pdf=from_pdf,
        on_error=on_error,
        lexical_graph_config=lexical_graph_config,
    )

    # Run the knowledge graph building process with text input
    await kg_builder_text.run_async(text=harry_potter_text)

    # check the content of the graph:
    # check lexical graph content
    records, _, _ = driver.execute_query("MATCH (start:chunkNodeLabel) RETURN start")
    assert len(records) == 1

    # check entity -> chunk relationships
    records, _, _ = driver.execute_query(
        "MATCH (chunk:chunkNodeLabel)<-[rel:FROM_CHUNK]-(entity:__Entity__) RETURN chunk, rel, entity"
    )
    assert len(records) == 2  # only two persons


@pytest.mark.asyncio
@pytest.mark.usefixtures("setup_neo4j_for_kg_construction")
async def test_pipeline_builder_two_documents(
    harry_potter_text_part1: str,
    harry_potter_text_part2: str,
    llm: MagicMock,
    embedder: MagicMock,
    driver: neo4j.Driver,
) -> None:
    """When everything works as expected, extracted entities, relations and text
    chunks must be in the DB
    """
    driver.execute_query("MATCH (n) DETACH DELETE n")
    embedder.async_embed_query.return_value = [1, 2, 3]
    llm.ainvoke.side_effect = [
        # first document
        # first chunk
        LLMResponse(
            content="""{
                        "nodes": [
                            {
                                "id": "0",
                                "label": "Person",
                                "properties": {
                                    "name": "Harry Potter"
                                }
                            },
                        ],
                        "relationships": []
                    }"""
        ),
        # second chunk
        LLMResponse(content='{"nodes": [], "relationships": []}'),
        # second document
        # first chunk
        LLMResponse(
            content="""{
                        "nodes": [
                            {
                                "id": "0",
                                "label": "Person",
                                "properties": {
                                    "name": "Hermione Granger"
                                }
                            },
                        ],
                        "relationships": []
                    }"""
        ),
        # second chunk
        LLMResponse(content='{"nodes": [], "relationships": []}'),
    ]

    # Create an instance of the SimpleKGPipeline
    kg_builder_text = SimpleKGPipeline(
        llm=llm,
        driver=driver,
        embedder=embedder,
        from_pdf=False,
        # provide minimal schema to bypass automatic schema extraction
        entities=["Person"],
        # in order to have 2 chunks:
        text_splitter=FixedSizeSplitter(chunk_size=400, chunk_overlap=5),
    )

    # Run the knowledge graph building process with text input
    await kg_builder_text.run_async(text=harry_potter_text_part1)
    await kg_builder_text.run_async(text=harry_potter_text_part2)

    # check graph content
    # check lexical graph content
    records, _, _ = driver.execute_query(
        "MATCH (start:Chunk)-[rel:NEXT_CHUNK]->(end:Chunk) RETURN start, rel, end"
    )
    assert len(records) == 2  # one for each run

    # check entity -> chunk relationships
    records, _, _ = driver.execute_query(
        "MATCH (chunk:Chunk)<-[rel:FROM_CHUNK]-(entity:__Entity__) RETURN chunk, rel, entity"
    )
    assert len(records) == 2  # two entities according to mocked LLMResponse


@pytest.mark.asyncio
@pytest.mark.usefixtures("setup_neo4j_for_kg_construction")
async def test_pipeline_builder_same_document_two_runs(
    harry_potter_text_part1: str,
    llm: MagicMock,
    embedder: MagicMock,
    driver: neo4j.Driver,
) -> None:
    """When everything works as expected, extracted entities, relations and text
    chunks must be in the DB
    """
    driver.execute_query("MATCH (n) DETACH DELETE n")
    embedder.async_embed_query.return_value = [1, 2, 3]
    llm.ainvoke.side_effect = [
        # first run
        # first chunk
        LLMResponse(
            content="""{
                        "nodes": [
                            {
                                "id": "0",
                                "label": "Person",
                                "properties": {
                                    "name": "Harry Potter"
                                }
                            },
                        ],
                        "relationships": []
                    }"""
        ),
        # second chunk
        LLMResponse(content='{"nodes": [], "relationships": []}'),
        # second run
        # first chunk
        LLMResponse(
            content="""{
                        "nodes": [
                            {
                                "id": "0",
                                "label": "Person",
                                "properties": {
                                    "name": "Harry Potter"
                                }
                            },
                        ],
                        "relationships": []
                    }"""
        ),
        # second chunk
        LLMResponse(content='{"nodes": [], "relationships": []}'),
    ]

    # Create an instance of the SimpleKGPipeline
    kg_builder_text = SimpleKGPipeline(
        llm=llm,
        driver=driver,
        embedder=embedder,
        from_pdf=False,
        # provide minimal schema to bypass automatic schema extraction
        entities=["Person"],
        # in order to have 2 chunks:
        text_splitter=FixedSizeSplitter(chunk_size=400, chunk_overlap=5),
    )

    # Run the knowledge graph building process with text input
    await kg_builder_text.run_async(text=harry_potter_text_part1)
    await kg_builder_text.run_async(text=harry_potter_text_part1)

    # check lexical graph content
    records, _, _ = driver.execute_query(
        "MATCH (start:Chunk)-[rel:NEXT_CHUNK]->(end:Chunk) RETURN start, rel, end"
    )
    assert len(records) == 2  # one for each run

    # check entity -> chunk relationships
    records, _, _ = driver.execute_query(
        "MATCH (chunk:Chunk)<-[rel:FROM_CHUNK]-(entity:__Entity__) RETURN chunk, rel, entity"
    )
    assert len(records) == 2  # two entities according to mocked LLMResponse


@pytest.mark.asyncio
@pytest.mark.usefixtures("setup_neo4j_for_kg_construction")
async def test_pipeline_builder_with_automatic_schema_extraction(
    harry_potter_text_part1: str,
    llm: MagicMock,
    embedder: MagicMock,
    driver: neo4j.Driver,
) -> None:
    """Test pipeline with automatic schema extraction (no schema provided).
    This test verifies that the pipeline correctly handles automatic schema extraction.
    """
    driver.execute_query("MATCH (n) DETACH DELETE n")
    embedder.async_embed_query.return_value = [1, 2, 3]

    # set up mock LLM responses for both schema extraction and entity extraction
    llm.ainvoke.side_effect = [
        # first call - schema extraction response
        LLMResponse(
            content="""{
                "node_types": [
                    {
                        "label": "Person",
                        "description": "A character in the story",
                        "properties": [
                            {"name": "name", "type": "STRING"},
                            {"name": "age", "type": "INTEGER"}
                        ]
                    },
                    {
                        "label": "Location",
                        "description": "A place in the story",
                        "properties": [
                            {"name": "name", "type": "STRING"}
                        ]
                    }
                ],
                "relationship_types": [
                    {
                        "label": "LOCATED_AT",
                        "description": "Indicates where a person is located",
                        "properties": []
                    }
                ],
                "patterns": [
                    ["Person", "LOCATED_AT", "Location"]
                ]
            }"""
        ),
        # second call - entity extraction for first chunk
        LLMResponse(
            content="""{
                "nodes": [
                    {
                        "id": "0",
                        "label": "Person",
                        "properties": {
                            "name": "Harry Potter"
                        }
                    },
                    {
                        "id": "1",
                        "label": "Location",
                        "properties": {
                            "name": "Hogwarts"
                        }
                    }
                ],
                "relationships": [
                    {
                        "type": "LOCATED_AT",
                        "start_node_id": "0",
                        "end_node_id": "1"
                    }
                ]
            }"""
        ),
        # third call - entity extraction for second chunk (if text is split)
        LLMResponse(content='{"nodes": [], "relationships": []}'),
    ]

    # create an instance of the SimpleKGPipeline with NO schema provided
    kg_builder_text = SimpleKGPipeline(
        llm=llm,
        driver=driver,
        embedder=embedder,
        from_pdf=False,
        # use smaller chunk size to ensure we have at least 2 chunks
        text_splitter=FixedSizeSplitter(chunk_size=400, chunk_overlap=5),
    )

    # run the knowledge graph building process with text input
    await kg_builder_text.run_async(text=harry_potter_text_part1)

    # verify LLM was called for schema extraction
    assert llm.ainvoke.call_count >= 2

    # verify entities were created
    records, _, _ = driver.execute_query("MATCH (n:Person) RETURN n")
    assert len(records) == 1

    # verify locations were created
    records, _, _ = driver.execute_query("MATCH (n:Location) RETURN n")
    assert len(records) == 1

    # verify relationships were created
    records, _, _ = driver.execute_query(
        "MATCH (p:Person)-[r:LOCATED_AT]->(l:Location) RETURN p, r, l"
    )
    assert len(records) == 1

    # verify chunks and relationships to entities
    records, _, _ = driver.execute_query(
        "MATCH (c:Chunk)<-[:FROM_CHUNK]-(e) RETURN c, e"
    )
    assert len(records) >= 1
